"""Test the HTML builder and check output against XPath."""

from __future__ import annotations

import os
import posixpath
import re
from typing import TYPE_CHECKING

import pytest

from sphinx._cli.util.errors import strip_escape_sequences
from sphinx.builders.html import (
    StandaloneHTMLBuilder,
    validate_html_extra_path,
    validate_html_static_path,
)
from sphinx.errors import ConfigError
from sphinx.testing.util import etree_parse
from sphinx.util.inventory import InventoryFile, _InventoryItem

from tests.test_builders.xpath_data import FIGURE_CAPTION
from tests.test_builders.xpath_util import check_xpath

if TYPE_CHECKING:
    from typing import Any

    from sphinx.testing.util import SphinxTestApp


def test_html_sidebars_error(make_app, tmp_path):
    (tmp_path / 'conf.py').touch()
    with pytest.raises(
        ConfigError,
        match="Values in 'html_sidebars' must be a list of strings. "
        "At least one pattern has a string value: 'index'. "
        r"Change to `html_sidebars = \{'index': \['searchbox.html'\]\}`.",
    ):
        make_app(
            'html',
            srcdir=tmp_path,
            confoverrides={'html_sidebars': {'index': 'searchbox.html'}},
        )


def test_html4_error(make_app, tmp_path):
    (tmp_path / 'conf.py').touch()
    with pytest.raises(
        ConfigError,
        match='HTML 4 is no longer supported by Sphinx',
    ):
        make_app(
            'html',
            srcdir=tmp_path,
            confoverrides={'html4_writer': True},
        )


@pytest.mark.parametrize(
    ('fname', 'path', 'check'),
    [
        ('index.html', ".//div[@class='citation']/span", r'Ref1'),
        ('index.html', ".//div[@class='citation']/span", r'Ref_1'),
        (
            'footnote.html',
            ".//a[@class='footnote-reference brackets'][@href='#id9'][@id='id1']",
            r'1',
        ),
        (
            'footnote.html',
            ".//a[@class='footnote-reference brackets'][@href='#id10'][@id='id2']",
            r'2',
        ),
        (
            'footnote.html',
            ".//a[@class='footnote-reference brackets'][@href='#foo'][@id='id3']",
            r'3',
        ),
        (
            'footnote.html',
            ".//a[@class='reference internal'][@href='#bar'][@id='id4']/span",
            r'\[bar\]',
        ),
        (
            'footnote.html',
            ".//a[@class='reference internal'][@href='#baz-qux'][@id='id5']/span",
            r'\[baz_qux\]',
        ),
        (
            'footnote.html',
            ".//a[@class='footnote-reference brackets'][@href='#id11'][@id='id6']",
            r'4',
        ),
        (
            'footnote.html',
            ".//a[@class='footnote-reference brackets'][@href='#id12'][@id='id7']",
            r'5',
        ),
        (
            'footnote.html',
            ".//aside[@class='footnote brackets']/span/a[@href='#id1']",
            r'1',
        ),
        (
            'footnote.html',
            ".//aside[@class='footnote brackets']/span/a[@href='#id2']",
            r'2',
        ),
        (
            'footnote.html',
            ".//aside[@class='footnote brackets']/span/a[@href='#id3']",
            r'3',
        ),
        ('footnote.html', ".//div[@class='citation']/span/a[@href='#id4']", r'bar'),
        ('footnote.html', ".//div[@class='citation']/span/a[@href='#id5']", r'baz_qux'),
        (
            'footnote.html',
            ".//aside[@class='footnote brackets']/span/a[@href='#id6']",
            r'4',
        ),
        (
            'footnote.html',
            ".//aside[@class='footnote brackets']/span/a[@href='#id7']",
            r'5',
        ),
        (
            'footnote.html',
            ".//aside[@class='footnote brackets']/span/a[@href='#id8']",
            r'6',
        ),
    ],
)
@pytest.mark.sphinx('html', testroot='root')
@pytest.mark.test_params(shared_result='test_build_html_output_docutils18')
def test_docutils_output(app, cached_etree_parse, fname, path, check):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check)


@pytest.mark.sphinx(
    'html',
    testroot='root',
    parallel=2,
)
def test_html_parallel(app: SphinxTestApp) -> None:
    app.build()


@pytest.mark.sphinx('html', testroot='build-html-translator')
def test_html_translator(app: SphinxTestApp) -> None:
    app.build()
    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    assert app.builder.docwriter.visitor.depart_with_node == 10


@pytest.mark.parametrize(
    'expect',
    [
        (FIGURE_CAPTION + "//span[@class='caption-number']", 'Fig. 1', True),
        (FIGURE_CAPTION + "//span[@class='caption-number']", 'Fig. 2', True),
        (FIGURE_CAPTION + "//span[@class='caption-number']", 'Fig. 3', True),
        (".//div//span[@class='caption-number']", 'No.1 ', True),
        (".//div//span[@class='caption-number']", 'No.2 ', True),
        ('.//li/p/a/span', 'Fig. 1', True),
        ('.//li/p/a/span', 'Fig. 2', True),
        ('.//li/p/a/span', 'Fig. 3', True),
        ('.//li/p/a/span', 'No.1', True),
        ('.//li/p/a/span', 'No.2', True),
    ],
)
@pytest.mark.sphinx(
    'html',
    testroot='add_enumerable_node',
    srcdir='test_enumerable_node',
)
def test_enumerable_node(app, cached_etree_parse, expect):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / 'index.html'), 'index.html', *expect)


@pytest.mark.sphinx(
    'html',
    testroot='basic',
    confoverrides={'html_copy_source': False},
)
def test_html_copy_source(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    assert not (app.outdir / '_sources' / 'index.rst.txt').exists()


@pytest.mark.sphinx(
    'html',
    testroot='basic',
    confoverrides={'html_sourcelink_suffix': '.txt'},
)
def test_html_sourcelink_suffix(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    assert (app.outdir / '_sources' / 'index.rst.txt').exists()


@pytest.mark.sphinx(
    'html',
    testroot='basic',
    confoverrides={'html_sourcelink_suffix': '.rst'},
)
def test_html_sourcelink_suffix_same(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    assert (app.outdir / '_sources' / 'index.rst').exists()


@pytest.mark.sphinx(
    'html',
    testroot='basic',
    confoverrides={'html_sourcelink_suffix': ''},
)
def test_html_sourcelink_suffix_empty(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    assert (app.outdir / '_sources' / 'index.rst').exists()


@pytest.mark.sphinx('html', testroot='html_entity')
def test_html_entity(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    valid_entities = {'amp', 'lt', 'gt', 'quot', 'apos'}
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    for entity in re.findall(r'&([a-z]+);', content, re.MULTILINE):
        assert entity not in valid_entities


@pytest.mark.sphinx('html', testroot='basic')
def test_html_inventory(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    with app.outdir.joinpath('objects.inv').open('rb') as f:
        invdata = InventoryFile.load(f, 'https://www.google.com', posixpath.join)

    assert set(invdata.keys()) == {'std:label', 'std:doc'}
    assert set(invdata['std:label'].keys()) == {
        'modindex',
        'py-modindex',
        'genindex',
        'search',
    }
    assert invdata['std:label']['modindex'] == _InventoryItem(
        project_name='Project name not set',
        project_version='',
        uri='https://www.google.com/py-modindex.html',
        display_name='Module Index',
    )
    assert invdata['std:label']['py-modindex'] == _InventoryItem(
        project_name='Project name not set',
        project_version='',
        uri='https://www.google.com/py-modindex.html',
        display_name='Python Module Index',
    )
    assert invdata['std:label']['genindex'] == _InventoryItem(
        project_name='Project name not set',
        project_version='',
        uri='https://www.google.com/genindex.html',
        display_name='Index',
    )
    assert invdata['std:label']['search'] == _InventoryItem(
        project_name='Project name not set',
        project_version='',
        uri='https://www.google.com/search.html',
        display_name='Search Page',
    )
    assert set(invdata['std:doc'].keys()) == {'index'}
    assert invdata['std:doc']['index'] == _InventoryItem(
        project_name='Project name not set',
        project_version='',
        uri='https://www.google.com/index.html',
        display_name='The basic Sphinx documentation for testing',
    )


@pytest.mark.usefixtures('_http_teapot')
@pytest.mark.sphinx(
    'html',
    testroot='images',
    confoverrides={'html_sourcelink_suffix': ''},
)
def test_html_anchor_for_figure(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert (
        '<figcaption>\n<p><span class="caption-text">The caption of pic</span>'
        '<a class="headerlink" href="#id1" title="Link to this image">¶</a></p>\n</figcaption>'
    ) in content


@pytest.mark.sphinx('html', testroot='directives-raw')
def test_html_raw_directive(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    result = (app.outdir / 'index.html').read_text(encoding='utf8')

    # standard case
    assert 'standalone raw directive (HTML)' in result
    assert 'standalone raw directive (LaTeX)' not in result

    # with substitution
    assert '<p>HTML: abc def ghi</p>' in result
    assert '<p>LaTeX: abc  ghi</p>' in result


@pytest.mark.parametrize(
    'expect',
    [
        (".//link[@href='_static/persistent.css'][@rel='stylesheet']", '', True),
        (
            ".//link[@href='_static/default.css'][@rel='stylesheet'][@title='Default']",
            '',
            True,
        ),
        (
            ".//link[@href='_static/alternate1.css']"
            "[@rel='alternate stylesheet']"
            "[@title='Alternate']",
            '',
            True,
        ),
        (
            ".//link[@href='_static/alternate2.css'][@rel='alternate stylesheet']",
            '',
            True,
        ),
        (
            ".//link[@href='_static/more_persistent.css'][@rel='stylesheet']",
            '',
            True,
        ),
        (
            ".//link[@href='_static/more_default.css']"
            "[@rel='stylesheet']"
            "[@title='Default']",
            '',
            True,
        ),
        (
            ".//link[@href='_static/more_alternate1.css']"
            "[@rel='alternate stylesheet']"
            "[@title='Alternate']",
            '',
            True,
        ),
        (
            ".//link[@href='_static/more_alternate2.css'][@rel='alternate stylesheet']",
            '',
            True,
        ),
    ],
)
@pytest.mark.sphinx('html', testroot='stylesheets')
def test_alternate_stylesheets(app, cached_etree_parse, expect):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / 'index.html'), 'index.html', *expect)


@pytest.mark.sphinx('html', testroot='html_style')
def test_html_style(app: SphinxTestApp) -> None:
    app.build()
    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert (
        '<link rel="stylesheet" type="text/css" href="_static/default.css" />'
    ) in result
    assert (
        '<link rel="stylesheet" type="text/css" href="_static/alabaster.css" />'
    ) not in result


@pytest.mark.sphinx(
    'html',
    testroot='basic',
    confoverrides={
        'html_sidebars': {
            '**': [
                'localtoc.html',
                'searchfield.html',
                'sourcelink.html',
            ]
        }
    },
)
def test_html_sidebar(app: SphinxTestApp) -> None:
    ctx: dict[str, Any] = {}

    # default for alabaster
    app.build(force_all=True)
    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    # layout.html
    assert '<div class="sphinxsidebar" role="navigation" aria-label="Main">' in result
    assert '<h1>The basic Sphinx documentation for testing' in result
    assert '<h3>Navigation</h3>' in result
    # localtoc.html
    assert '<h3><a href="#">Table of Contents</a></h3>' in result
    # searchfield.html
    assert '<div class="searchformwrapper">' in result
    # sourcelink.html
    assert '<h3>This Page</h3>' in result

    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    app.builder.add_sidebars('index', ctx)
    assert ctx['sidebars'] == [
        'localtoc.html',
        'searchfield.html',
        'sourcelink.html',
    ]

    # only sourcelink.html
    app.config.html_sidebars = {'**': ['sourcelink.html']}
    app.build(force_all=True)
    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    # layout.html
    assert '<div class="sphinxsidebar" role="navigation" aria-label="Main">' in result
    assert '<h1>The basic Sphinx documentation for testing' in result
    assert '<h3>Navigation</h3>' in result
    # localtoc.html
    assert '<h3><a href="#">Table of Contents</a></h3>' not in result
    # searchfield.html
    assert '<div class="searchformwrapper">' not in result
    # sourcelink.html
    assert '<h3>This Page</h3>' in result

    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    app.builder.add_sidebars('index', ctx)
    assert ctx['sidebars'] == ['sourcelink.html']

    # no sidebars
    app.config.html_sidebars = {'**': []}
    app.build(force_all=True)
    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    # layout.html
    assert (
        '<div class="sphinxsidebar" role="navigation" aria-label="Main">'
    ) not in result
    assert '<h1>The basic Sphinx documentation for testing' in result
    assert '<h3>Navigation</h3>' in result
    # localtoc.html
    assert '<h3><a href="#">Table of Contents</a></h3>' not in result
    # searchfield.html
    assert '<div class="searchformwrapper">' not in result
    # sourcelink.html
    assert '<h3>This Page</h3>' not in result

    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    app.builder.add_sidebars('index', ctx)
    assert ctx['sidebars'] == []


@pytest.mark.parametrize(
    ('fname', 'expect'),
    [
        ('index.html', (".//h1/em/a[@href='https://example.com/cp.1']", '', True)),
        ('index.html', (".//em/a[@href='https://example.com/man.1']", '', True)),
        ('index.html', (".//em/a[@href='https://example.com/ls.1']", '', True)),
        ('index.html', (".//em/a[@href='https://example.com/sphinx.']", '', True)),
    ],
)
@pytest.mark.sphinx(
    'html',
    testroot='manpage_url',
    confoverrides={'manpages_url': 'https://example.com/{page}.{section}'},
)
@pytest.mark.test_params(shared_result='test_build_html_manpage_url')
def test_html_manpage(app, cached_etree_parse, fname, expect):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, *expect)


@pytest.mark.sphinx(
    'html',
    testroot='toctree-glob',
    confoverrides={'html_baseurl': 'https://example.com/'},
)
def test_html_baseurl(app: SphinxTestApp) -> None:
    app.build()

    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<link rel="canonical" href="https://example.com/index.html" />' in result

    result = (app.outdir / 'qux' / 'index.html').read_text(encoding='utf8')
    assert (
        '<link rel="canonical" href="https://example.com/qux/index.html" />'
    ) in result


@pytest.mark.sphinx(
    'html',
    testroot='toctree-glob',
    confoverrides={
        'html_baseurl': 'https://example.com/subdir',
        'html_file_suffix': '.htm',
    },
)
def test_html_baseurl_and_html_file_suffix(app: SphinxTestApp) -> None:
    app.build()

    result = (app.outdir / 'index.htm').read_text(encoding='utf8')
    assert (
        '<link rel="canonical" href="https://example.com/subdir/index.htm" />'
    ) in result

    result = (app.outdir / 'qux' / 'index.htm').read_text(encoding='utf8')
    assert (
        '<link rel="canonical" href="https://example.com/subdir/qux/index.htm" />'
    ) in result


@pytest.mark.sphinx(
    'html',
    testroot='basic',
    srcdir='validate_html_extra_path',
)
def test_validate_html_extra_path(app: SphinxTestApp) -> None:
    (app.confdir / '_static').mkdir(parents=True, exist_ok=True)
    app.config.html_extra_path = [
        '/path/to/not_found',    # not found
        '_static',
        app.outdir,              # outdir
        app.outdir / '_static',  # inside outdir
    ]  # fmt: skip

    validate_html_extra_path(app, app.config)
    assert app.config.html_extra_path == ['_static']

    warnings = strip_escape_sequences(app.warning.getvalue()).splitlines()
    assert "html_extra_path entry '/path/to/not_found' does not exist" in warnings[0]
    assert warnings[1].endswith(' is placed inside outdir')
    assert warnings[2].endswith(' does not exist')
    assert len(warnings) == 3


@pytest.mark.sphinx(
    'html',
    testroot='basic',
    srcdir='validate_html_static_path',
)
def test_validate_html_static_path(app: SphinxTestApp) -> None:
    (app.confdir / '_static').mkdir(parents=True, exist_ok=True)
    app.config.html_static_path = [
        '/path/to/not_found',    # not found
        '_static',
        app.outdir,              # outdir
        app.outdir / '_static',  # inside outdir
    ]  # fmt: skip

    validate_html_static_path(app, app.config)
    assert app.config.html_static_path == ['_static']

    warnings = strip_escape_sequences(app.warning.getvalue()).splitlines()
    assert "html_static_path entry '/path/to/not_found' does not exist" in warnings[0]
    assert warnings[1].endswith(' is placed inside outdir')
    assert warnings[2].endswith(' does not exist')
    assert len(warnings) == 3


@pytest.mark.sphinx(
    'html',
    testroot='basic',
    confoverrides={'html_permalinks': False},
)
def test_html_permalink_disable(app: SphinxTestApp) -> None:
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    assert '<h1>The basic Sphinx documentation for testing</h1>' in content


@pytest.mark.sphinx(
    'html',
    testroot='basic',
    confoverrides={'html_permalinks_icon': '<span>[PERMALINK]</span>'},
)
def test_html_permalink_icon(app: SphinxTestApp) -> None:
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    assert (
        '<h1>The basic Sphinx documentation for testing<a class="headerlink" '
        'href="#the-basic-sphinx-documentation-for-testing" '
        'title="Link to this heading"><span>[PERMALINK]</span></a></h1>'
    ) in content


@pytest.mark.sphinx('html', testroot='html_signaturereturn_icon')
def test_html_signaturereturn_icon(app: SphinxTestApp) -> None:
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    assert '<span class="sig-return-icon">&#x2192;</span>' in content


@pytest.mark.sphinx(
    'html',
    testroot='root',
    srcdir=os.urandom(4).hex(),
)
def test_html_remove_sources_before_write_gh_issue_10786(app: SphinxTestApp) -> None:
    # See: https://github.com/sphinx-doc/sphinx/issues/10786
    target = app.srcdir / 'img.png'

    def handler(app):
        assert target.exists()
        target.unlink()
        return []

    app.connect('html-collect-pages', handler)
    assert target.exists()
    app.build()
    assert not target.exists()

    ws = strip_escape_sequences(app.warning.getvalue()).splitlines()
    assert len(ws) >= 1

    file = os.fsdecode(target)
    assert f"WARNING: cannot copy image file '{file}': {file} does not exist" == ws[-1]


@pytest.mark.sphinx(
    'html',
    testroot='domain-py-python_maximum_signature_line_length',
    confoverrides={'python_maximum_signature_line_length': 1},
)
def test_html_pep_695_one_type_per_line(app, cached_etree_parse):
    app.build()
    fname = app.outdir / 'index.html'
    etree = cached_etree_parse(fname)

    class chk:
        def __init__(self, expect: str) -> None:
            self.expect = expect

        def __call__(self, nodes):
            assert len(nodes) == 1, nodes
            objnode = ''.join(nodes[0].itertext()).replace('\n\n', '')
            objnode = objnode.rstrip(chr(182))  # remove '¶' symbol
            objnode = objnode.strip('\n')  # remove surrounding new lines
            assert objnode == self.expect

    # each signature has a dangling ',' at the end of its parameters lists
    check_xpath(
        etree,
        fname,
        r'.//dt[@id="generic_foo"][1]',
        chk('generic_foo[\nT,\n]()'),
    )
    check_xpath(
        etree,
        fname,
        r'.//dt[@id="generic_bar"][1]',
        chk('generic_bar[\nT,\n](\nx: list[T],\n)'),
    )
    check_xpath(
        etree,
        fname,
        r'.//dt[@id="generic_ret"][1]',
        chk('generic_ret[\nR,\n]() → R'),
    )
    check_xpath(
        etree,
        fname,
        r'.//dt[@id="MyGenericClass"][1]',
        chk('class MyGenericClass[\nX,\n]'),
    )
    check_xpath(
        etree,
        fname,
        r'.//dt[@id="MyList"][1]',
        chk('class MyList[\nT,\n](list[T])'),
    )


@pytest.mark.sphinx(
    'html',
    testroot='domain-py-python_maximum_signature_line_length',
    confoverrides={
        'python_maximum_signature_line_length': 1,
        'python_trailing_comma_in_multi_line_signatures': False,
    },
)
def test_html_pep_695_trailing_comma_in_multi_line_signatures(
    app: SphinxTestApp,
) -> None:
    app.build()
    fname = app.outdir / 'index.html'
    etree = etree_parse(fname)

    class chk:
        def __init__(self, expect: str) -> None:
            self.expect = expect

        def __call__(self, nodes):
            assert len(nodes) == 1, nodes
            objnode = ''.join(nodes[0].itertext()).replace('\n\n', '')
            objnode = objnode.rstrip(chr(182))  # remove '¶' symbol
            objnode = objnode.strip('\n')  # remove surrounding new lines
            assert objnode == self.expect

    # each signature has a dangling ',' at the end of its parameters lists
    check_xpath(
        etree,
        fname,
        r'.//dt[@id="generic_foo"][1]',
        chk('generic_foo[\nT\n]()'),
    )
    check_xpath(
        etree,
        fname,
        r'.//dt[@id="generic_bar"][1]',
        chk('generic_bar[\nT\n](\nx: list[T]\n)'),
    )
    check_xpath(
        etree,
        fname,
        r'.//dt[@id="generic_ret"][1]',
        chk('generic_ret[\nR\n]() → R'),
    )
    check_xpath(
        etree,
        fname,
        r'.//dt[@id="MyGenericClass"][1]',
        chk('class MyGenericClass[\nX\n]'),
    )
    check_xpath(
        etree,
        fname,
        r'.//dt[@id="MyList"][1]',
        chk('class MyList[\nT\n](list[T])'),
    )


@pytest.mark.sphinx('html', testroot='directives-admonition-collapse')
def test_html_admonition_collapse(app: SphinxTestApp) -> None:
    app.build()
    fname = app.outdir / 'index.html'
    etree = etree_parse(fname)

    def _create_check(text: str, open: bool):  # type: ignore[no-untyped-def]
        def _check(els):
            assert len(els) == 1
            el = els[0]
            if open:
                assert el.attrib['open'] == 'open'
            else:
                assert 'open' not in el.attrib
            assert el.find('p').text == text

        return _check

    check_xpath(
        etree,
        fname,
        r'.//div[@class="standard admonition note"]//p',
        'This is a standard note.',
    )
    check_xpath(
        etree,
        fname,
        r'.//details[@class="admonition note"]',
        _create_check('This note is collapsible, and initially open by default.', True),
    )
    check_xpath(
        etree,
        fname,
        r'.//details[@class="admonition-example admonition"]',
        _create_check('This example is collapsible, and initially open.', True),
    )
    check_xpath(
        etree,
        fname,
        r'.//details[@class="admonition hint"]',
        _create_check('This hint is collapsible, but initially closed.', False),
    )



import sys
from pathlib import Path

sys.path.insert(0, str(Path.cwd().resolve()))

from docutils.parsers import Parser


class DummyTestParser(Parser):
    supported = ('dummy',)


extensions = ['source_parser']
source_suffix = {
    '.rst': 'restructuredtext',
    '.test': 'restructuredtext',
}
source_parsers = {
    '.test': DummyTestParser,
}



"""Test copyright year adjustment"""

from __future__ import annotations

import time
from typing import TYPE_CHECKING

import pytest

from sphinx.config import (
    Config,
    correct_copyright_year,
    evaluate_copyright_placeholders,
)
from sphinx.testing.util import SphinxTestApp

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from pathlib import Path

    from sphinx.testing.util import SphinxTestApp

LT = time.localtime()
LT_NEW = (2009, *LT[1:], LT.tm_zone, LT.tm_gmtoff)
LOCALTIME_2009 = type(LT)(LT_NEW)


@pytest.fixture(
    params=[
        # test with SOURCE_DATE_EPOCH unset: no modification
        (None, None),
        # test with post-2009 SOURCE_DATE_EPOCH set: copyright year should not be updated
        ('1293840000', 2011),
        ('1293839999', 2010),
        # test with pre-2009 SOURCE_DATE_EPOCH set: copyright year should be updated
        ('1199145600', 2008),
        ('1199145599', 2007),
    ],
)
def expect_date(
    request: pytest.FixtureRequest,
    monkeypatch: pytest.MonkeyPatch,
) -> Iterator[int | None]:
    sde, expect = request.param
    with monkeypatch.context() as m:
        m.setattr(time, 'localtime', lambda *a: LOCALTIME_2009)
        if sde:
            m.setenv('SOURCE_DATE_EPOCH', sde)
        else:
            m.delenv('SOURCE_DATE_EPOCH', raising=False)
        yield expect


def test_correct_year(expect_date: int | None) -> None:
    # test that copyright is substituted
    copyright_date = '2006-2009, Alice'
    cfg = Config({'copyright': copyright_date}, {})
    assert cfg.copyright == copyright_date
    correct_copyright_year(None, cfg)  # type: ignore[arg-type]
    if expect_date and expect_date <= LOCALTIME_2009.tm_year:
        assert cfg.copyright == f'2006-{expect_date}, Alice'
    else:
        assert cfg.copyright == copyright_date


def test_correct_year_space(expect_date: int | None) -> None:
    # test that copyright is substituted
    copyright_date = '2006-2009 Alice'
    cfg = Config({'copyright': copyright_date}, {})
    assert cfg.copyright == copyright_date
    correct_copyright_year(None, cfg)  # type: ignore[arg-type]
    if expect_date and expect_date <= LOCALTIME_2009.tm_year:
        assert cfg.copyright == f'2006-{expect_date} Alice'
    else:
        assert cfg.copyright == copyright_date


def test_correct_year_no_author(expect_date: int | None) -> None:
    # test that copyright is substituted
    copyright_date = '2006-2009'
    cfg = Config({'copyright': copyright_date}, {})
    assert cfg.copyright == copyright_date
    correct_copyright_year(None, cfg)  # type: ignore[arg-type]
    if expect_date and expect_date <= LOCALTIME_2009.tm_year:
        assert cfg.copyright == f'2006-{expect_date}'
    else:
        assert cfg.copyright == copyright_date


def test_correct_year_single(expect_date: int | None) -> None:
    # test that copyright is substituted
    copyright_date = '2009, Alice'
    cfg = Config({'copyright': copyright_date}, {})
    assert cfg.copyright == copyright_date
    correct_copyright_year(None, cfg)  # type: ignore[arg-type]
    if expect_date and expect_date <= LOCALTIME_2009.tm_year:
        assert cfg.copyright == f'{expect_date}, Alice'
    else:
        assert cfg.copyright == copyright_date


def test_correct_year_single_space(expect_date: int | None) -> None:
    # test that copyright is substituted
    copyright_date = '2009 Alice'
    cfg = Config({'copyright': copyright_date}, {})
    assert cfg.copyright == copyright_date
    correct_copyright_year(None, cfg)  # type: ignore[arg-type]
    if expect_date and expect_date <= LOCALTIME_2009.tm_year:
        assert cfg.copyright == f'{expect_date} Alice'
    else:
        assert cfg.copyright == copyright_date


def test_correct_year_single_no_author(expect_date: int | None) -> None:
    # test that copyright is substituted
    copyright_date = '2009'
    cfg = Config({'copyright': copyright_date}, {})
    assert cfg.copyright == copyright_date
    correct_copyright_year(None, cfg)  # type: ignore[arg-type]
    if expect_date and expect_date <= LOCALTIME_2009.tm_year:
        assert cfg.copyright == f'{expect_date}'
    else:
        assert cfg.copyright == copyright_date


def test_correct_year_placeholder(expect_date: int | None) -> None:
    # test that copyright is substituted
    copyright_date = '2006-%Y, Alice'
    cfg = Config({'copyright': copyright_date}, {})
    assert cfg.copyright == copyright_date
    evaluate_copyright_placeholders(None, cfg)  # type: ignore[arg-type]
    correct_copyright_year(None, cfg)  # type: ignore[arg-type]
    if expect_date and expect_date <= LOCALTIME_2009.tm_year:
        assert cfg.copyright == f'2006-{expect_date}, Alice'
    else:
        assert cfg.copyright == '2006-2009, Alice'


def test_correct_year_multi_line(expect_date: int | None) -> None:
    # test that copyright is substituted
    copyright_dates = (
        '2009',
        '2006-2009, Alice',
        '2010-2013, Bob',
        '2014-2017, Charlie',
        '2018-2021, David',
        '2022-2025, Eve',
    )
    cfg = Config({'copyright': copyright_dates}, {})
    assert cfg.copyright == copyright_dates
    correct_copyright_year(None, cfg)  # type: ignore[arg-type]
    if expect_date and expect_date <= LOCALTIME_2009.tm_year:
        assert cfg.copyright == (
            f'{expect_date}',
            f'2006-{expect_date}, Alice',
            # post 2009-dates aren't substituted
            '2010-2013, Bob',
            '2014-2017, Charlie',
            '2018-2021, David',
            '2022-2025, Eve',
        )
    else:
        assert cfg.copyright == copyright_dates


def test_correct_year_multi_line_all_formats(expect_date: int | None) -> None:
    # test that copyright is substituted
    copyright_dates = (
        '2009',
        '2009 Alice',
        '2009, Bob',
        '2006-2009',
        '2006-2009 Charlie',
        '2006-2009, David',
    )
    cfg = Config({'copyright': copyright_dates}, {})
    assert cfg.copyright == copyright_dates
    correct_copyright_year(None, cfg)  # type: ignore[arg-type]
    if expect_date and expect_date <= LOCALTIME_2009.tm_year:
        assert cfg.copyright == (
            f'{expect_date}',
            f'{expect_date} Alice',
            f'{expect_date}, Bob',
            f'2006-{expect_date}',
            f'2006-{expect_date} Charlie',
            f'2006-{expect_date}, David',
        )
    else:
        assert cfg.copyright == copyright_dates


def test_correct_year_multi_line_all_formats_placeholder(
    expect_date: int | None,
) -> None:
    # test that copyright is substituted
    copyright_dates = (
        '%Y',
        '%Y Alice',
        '%Y, Bob',
        '2006-%Y',
        '2006-%Y Charlie',
        '2006-%Y, David',
        # other format codes are left as-is
        '2006-%y, Eve',
        '%Y-%m-%d %H:%M:S %z, Francis',
    )
    cfg = Config({'copyright': copyright_dates}, {})
    assert cfg.copyright == copyright_dates
    evaluate_copyright_placeholders(None, cfg)  # type: ignore[arg-type]
    correct_copyright_year(None, cfg)  # type: ignore[arg-type]
    if expect_date and expect_date <= LOCALTIME_2009.tm_year:
        assert cfg.copyright == (
            f'{expect_date}',
            f'{expect_date} Alice',
            f'{expect_date}, Bob',
            f'2006-{expect_date}',
            f'2006-{expect_date} Charlie',
            f'2006-{expect_date}, David',
            '2006-%y, Eve',
            '2009-%m-%d %H:%M:S %z, Francis',
        )
    else:
        assert cfg.copyright == (
            '2009',
            '2009 Alice',
            '2009, Bob',
            '2006-2009',
            '2006-2009 Charlie',
            '2006-2009, David',
            '2006-%y, Eve',
            '2009-%m-%d %H:%M:S %z, Francis',
        )


def test_correct_year_app(
    expect_date: int | None,
    tmp_path: Path,
    make_app: Callable[..., SphinxTestApp],
) -> None:
    # integration test
    copyright_date = '2006-2009, Alice'
    (tmp_path / 'conf.py').touch()
    app = make_app(
        'dummy',
        srcdir=tmp_path,
        confoverrides={'copyright': copyright_date},
    )
    if expect_date and expect_date <= LOCALTIME_2009.tm_year:
        assert app.config.copyright == f'2006-{expect_date}, Alice'
    else:
        assert app.config.copyright == copyright_date



from functools import singledispatchmethod


class Foo:
    """docstring"""

    @singledispatchmethod
    @classmethod
    def class_meth(cls, arg, kwarg=None):
        """A class method for general use."""
        pass

    @class_meth.register(int)
    @class_meth.register(float)
    @classmethod
    def _class_meth_int(cls, arg, kwarg=None):
        """A class method for numbers."""
        pass

    @class_meth.register(str)
    @classmethod
    def _class_meth_str(cls, arg, kwarg=None):
        """A class method for str."""
        pass

    @class_meth.register
    @classmethod
    def _class_meth_dict(cls, arg: dict, kwarg=None):
        """A class method for dict."""
        # This function tests for specifying type through annotations
        pass



"""Test the autodoc extension.

This tests mainly the Documenters; the auto directives are tested in a test
source file translated by test_build.
"""

from __future__ import annotations

import inspect
import sys
import typing

import pytest

from tests.test_extensions.autodoc_util import do_autodoc

if typing.TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='ext-autodoc')
def test_empty_all(app: SphinxTestApp) -> None:
    options = {'members': None}
    actual = do_autodoc(app, 'module', 'target.empty_all', options)
    assert list(actual) == [
        '',
        '.. py:module:: target.empty_all',
        '',
        '   docsting of empty_all module.',
        '',
    ]


@pytest.mark.sphinx('html', testroot='ext-autodoc')
def test_automodule(app: SphinxTestApp) -> None:
    options = {'members': None}
    actual = do_autodoc(app, 'module', 'target.module', options)
    assert list(actual) == [
        '',
        '.. py:module:: target.module',
        '',
        '',
        '.. py:data:: annotated',
        '   :module: target.module',
        '   :type: int',
        '',
        '   docstring',
        '',
        '',
        '.. py:data:: documented',
        '   :module: target.module',
        '   :value: 1',
        '',
        '   docstring',
        '',
    ]


@pytest.mark.sphinx('html', testroot='ext-autodoc')
def test_automodule_undoc_members(app: SphinxTestApp) -> None:
    options = {
        'members': None,
        'undoc-members': None,
    }
    actual = do_autodoc(app, 'module', 'target.module', options)
    assert list(actual) == [
        '',
        '.. py:module:: target.module',
        '',
        '',
        '.. py:data:: annotated',
        '   :module: target.module',
        '   :type: int',
        '',
        '   docstring',
        '',
        '',
        '.. py:data:: documented',
        '   :module: target.module',
        '   :value: 1',
        '',
        '   docstring',
        '',
        '',
        '.. py:data:: undoc_annotated',
        '   :module: target.module',
        '   :type: int',
        '',
    ]


@pytest.mark.sphinx('html', testroot='ext-autodoc')
def test_automodule_special_members(app: SphinxTestApp) -> None:
    options = {
        'members': None,
        'special-members': None,
    }
    actual = do_autodoc(app, 'module', 'target.module', options)
    assert list(actual) == [
        '',
        '.. py:module:: target.module',
        '',
        '',
        '.. py:data:: __documented_special__',
        '   :module: target.module',
        '   :value: 1',
        '',
        '   docstring',
        '',
        '',
        '.. py:data:: annotated',
        '   :module: target.module',
        '   :type: int',
        '',
        '   docstring',
        '',
        '',
        '.. py:data:: documented',
        '   :module: target.module',
        '   :value: 1',
        '',
        '   docstring',
        '',
    ]


@pytest.mark.sphinx('html', testroot='ext-autodoc')
def test_automodule_inherited_members(app: SphinxTestApp) -> None:
    options = {
        'members': None,
        'undoc-members': None,
        'inherited-members': 'Base, list',
    }
    actual = do_autodoc(app, 'module', 'target.inheritance', options)
    assert list(actual) == [
        '',
        '.. py:module:: target.inheritance',
        '',
        '',
        '.. py:class:: AnotherBase()',
        '   :module: target.inheritance',
        '',
        '',
        '   .. py:method:: AnotherBase.another_inheritedmeth()',
        '      :module: target.inheritance',
        '',
        '      Another inherited function.',
        '',
        '',
        '.. py:class:: Base()',
        '   :module: target.inheritance',
        '',
        '',
        '   .. py:attribute:: Base.inheritedattr',
        '      :module: target.inheritance',
        '      :value: None',
        '',
        '      docstring',
        '',
        '',
        '   .. py:method:: Base.inheritedclassmeth()',
        '      :module: target.inheritance',
        '      :classmethod:',
        '',
        '      Inherited class method.',
        '',
        '',
        '   .. py:method:: Base.inheritedmeth()',
        '      :module: target.inheritance',
        '',
        '      Inherited function.',
        '',
        '',
        '   .. py:method:: Base.inheritedstaticmeth(cls)',
        '      :module: target.inheritance',
        '      :staticmethod:',
        '',
        '      Inherited static method.',
        '',
        '',
        '.. py:class:: Derived()',
        '   :module: target.inheritance',
        '',
        '',
        '   .. py:method:: Derived.another_inheritedmeth()',
        '      :module: target.inheritance',
        '',
        '      Another inherited function.',
        '',
        '',
        '   .. py:method:: Derived.inheritedmeth()',
        '      :module: target.inheritance',
        '',
        '      Inherited function.',
        '',
        '',
        '.. py:class:: MyList(iterable=(), /)',
        '   :module: target.inheritance',
        '',
        '',
        '   .. py:method:: MyList.meth()',
        '      :module: target.inheritance',
        '',
        '      docstring',
        '',
    ]


@pytest.mark.sphinx(
    'html',
    testroot='ext-autodoc',
    confoverrides={
        'autodoc_mock_imports': [
            'missing_module',
            'missing_package1',
            'missing_package2',
            'missing_package3',
            'sphinx.missing_module4',
        ]
    },
)
@pytest.mark.usefixtures('rollback_sysmodules')
def test_subclass_of_mocked_object(app: SphinxTestApp) -> None:
    from sphinx.ext.autodoc.mock import _MockObject

    sys.modules.pop('target', None)  # unload target module to clear the module cache

    options = {'members': None}
    actual = do_autodoc(app, 'module', 'target.need_mocks', options)
    # ``typing.Any`` is not available at runtime on ``_MockObject.__new__``
    assert '.. py:class:: Inherited(*args: Any, **kwargs: Any)' in actual

    # make ``typing.Any`` available at runtime on ``_MockObject.__new__``
    sig = inspect.signature(_MockObject.__new__)
    parameters = sig.parameters.copy()
    for name in ('args', 'kwargs'):
        parameters[name] = parameters[name].replace(annotation=typing.Any)
    sig = sig.replace(parameters=tuple(parameters.values()))
    _MockObject.__new__.__signature__ = sig  # type: ignore[attr-defined]

    options = {'members': None}
    actual = do_autodoc(app, 'module', 'target.need_mocks', options)
    assert (
        '.. py:class:: Inherited(*args: ~typing.Any, **kwargs: ~typing.Any)'
    ) in actual



"""Tests the std domain"""

from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import pytest
from docutils import nodes
from docutils.nodes import definition, definition_list, definition_list_item, term

from sphinx import addnodes
from sphinx.addnodes import (
    desc,
    desc_addname,
    desc_content,
    desc_name,
    desc_signature,
    glossary,
    index,
    pending_xref,
)
from sphinx.domains.std import StandardDomain
from sphinx.testing import restructuredtext
from sphinx.testing.util import assert_node

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


def test_process_doc_handle_figure_caption() -> None:
    env = mock.Mock(domaindata={})
    env._registry.enumerable_nodes = {}
    figure_node = nodes.figure(
        '',
        nodes.caption('caption text', 'caption text'),
    )
    document = mock.Mock(
        nametypes={'testname': True},
        nameids={'testname': 'testid'},
        ids={'testid': figure_node},
        citation_refs={},
    )
    document.findall.return_value = []

    domain = StandardDomain(env)
    if 'testname' in domain.data['labels']:
        del domain.data['labels']['testname']
    domain.process_doc(env, 'testdoc', document)
    assert 'testname' in domain.data['labels']
    assert domain.data['labels']['testname'] == ('testdoc', 'testid', 'caption text')


def test_process_doc_handle_table_title() -> None:
    env = mock.Mock(domaindata={})
    env._registry.enumerable_nodes = {}
    table_node = nodes.table(
        '',
        nodes.title('title text', 'title text'),
    )
    document = mock.Mock(
        nametypes={'testname': True},
        nameids={'testname': 'testid'},
        ids={'testid': table_node},
        citation_refs={},
    )
    document.findall.return_value = []

    domain = StandardDomain(env)
    if 'testname' in domain.data['labels']:
        del domain.data['labels']['testname']
    domain.process_doc(env, 'testdoc', document)
    assert 'testname' in domain.data['labels']
    assert domain.data['labels']['testname'] == ('testdoc', 'testid', 'title text')


def test_get_full_qualified_name() -> None:
    env = mock.Mock(domaindata={})
    env._registry.enumerable_nodes = {}
    domain = StandardDomain(env)

    # normal references
    node = nodes.reference()
    assert domain.get_full_qualified_name(node) is None

    # simple reference to options
    node = nodes.reference(reftype='option', reftarget='-l')
    assert domain.get_full_qualified_name(node) is None

    # options with std:program context
    kwargs = {'std:program': 'ls'}
    node = nodes.reference(reftype='option', reftarget='-l', **kwargs)
    assert domain.get_full_qualified_name(node) == 'ls.-l'


@pytest.mark.sphinx('html', testroot='_blank')
def test_cmd_option_with_optional_value(app: SphinxTestApp) -> None:
    text = '.. option:: -j[=N]'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            index,
            [
                desc,
                (
                    [desc_signature, ([desc_name, '-j'], [desc_addname, '[=N]'])],
                    [desc_content, ()],
                ),
            ],
        ),
    )
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('pair', 'command line option; -j', 'cmdoption-j', '', None)],
    )

    objects = list(app.env.domains.standard_domain.get_objects())
    assert ('-j', '-j', 'cmdoption', 'index', 'cmdoption-j', 1) in objects


@pytest.mark.sphinx('html', testroot='_blank')
def test_cmd_option_starting_with_bracket(app: SphinxTestApp) -> None:
    text = '.. option:: [enable=]PATTERN'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            index,
            [
                desc,
                (
                    [
                        desc_signature,
                        ([desc_name, '[enable'], [desc_addname, '=]PATTERN']),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )
    objects = list(app.env.domains.standard_domain.get_objects())
    assert (
        '[enable',
        '[enable',
        'cmdoption',
        'index',
        'cmdoption-arg-enable',
        1,
    ) in objects


@pytest.mark.sphinx('html', testroot='_blank')
def test_glossary(app):
    text = (
        '.. glossary::\n'
        '\n'
        '   term1\n'
        '   TERM2\n'
        '       description\n'
        '\n'
        '   term3 : classifier\n'
        '       description\n'
        '       description\n'
        '\n'
        '   term4 : class1 : class2\n'
        '       description\n'
    )

    # doctree
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            [
                glossary,
                definition_list,
                (
                    [
                        definition_list_item,
                        (
                            [term, ('term1', index)],
                            [term, ('TERM2', index)],
                            definition,
                        ),
                    ],
                    [definition_list_item, ([term, ('term3', index)], definition)],
                    [definition_list_item, ([term, ('term4', index)], definition)],
                ),
            ],
        ),
    )
    assert_node(
        doctree[0][0][0][0][1],
        entries=[('single', 'term1', 'term-term1', 'main', None)],
    )
    assert_node(
        doctree[0][0][0][1][1],
        entries=[('single', 'TERM2', 'term-TERM2', 'main', None)],
    )
    assert_node(doctree[0][0][0][2], [definition, nodes.paragraph, 'description'])
    assert_node(
        doctree[0][0][1][0][1],
        entries=[('single', 'term3', 'term-term3', 'main', 'classifier')],
    )
    assert_node(
        doctree[0][0][1][1],
        [definition, nodes.paragraph, 'description\ndescription'],
    )
    assert_node(
        doctree[0][0][2][0][1],
        entries=[('single', 'term4', 'term-term4', 'main', 'class1')],
    )
    assert_node(doctree[0][0][2][1], [nodes.definition, nodes.paragraph, 'description'])

    # index
    domain = app.env.domains.standard_domain
    objects = list(domain.get_objects())
    assert ('term1', 'term1', 'term', 'index', 'term-term1', -1) in objects
    assert ('TERM2', 'TERM2', 'term', 'index', 'term-TERM2', -1) in objects
    assert ('term3', 'term3', 'term', 'index', 'term-term3', -1) in objects
    assert ('term4', 'term4', 'term', 'index', 'term-term4', -1) in objects

    # term reference (case sensitive)
    refnode = domain.resolve_xref(
        app.env,
        'index',
        app.builder,
        'term',
        'term1',
        pending_xref(),
        nodes.paragraph(),
    )
    assert_node(refnode, nodes.reference, refid='term-term1')

    # term reference (case insensitive)
    refnode = domain.resolve_xref(
        app.env,
        'index',
        app.builder,
        'term',
        'term2',
        pending_xref(),
        nodes.paragraph(),
    )
    assert_node(refnode, nodes.reference, refid='term-TERM2')


@pytest.mark.sphinx('html', testroot='_blank')
def test_glossary_warning(app: SphinxTestApp) -> None:
    # empty line between terms
    text = '.. glossary::\n\n   term1\n\n   term2\n'
    restructuredtext.parse(app, text, 'case1')
    assert (
        'case1.rst:4: WARNING: glossary terms must not be separated by empty lines'
    ) in app.warning.getvalue()

    # glossary starts with indented item
    text = '.. glossary::\n\n       description\n   term\n'
    restructuredtext.parse(app, text, 'case2')
    assert (
        'case2.rst:3: WARNING: glossary term must be preceded by empty line'
    ) in app.warning.getvalue()

    # empty line between terms
    text = '.. glossary::\n\n   term1\n       description\n   term2\n'
    restructuredtext.parse(app, text, 'case3')
    assert (
        'case3.rst:4: WARNING: glossary term must be preceded by empty line'
    ) in app.warning.getvalue()

    # duplicated terms
    text = '.. glossary::\n\n   term-case4\n   term-case4\n'
    restructuredtext.parse(app, text, 'case4')
    assert (
        'case4.rst:3: WARNING: duplicate term description of term-case4, '
        'other instance in case4'
    ) in app.warning.getvalue()


@pytest.mark.sphinx('html', testroot='_blank')
def test_glossary_comment(app):
    text = (
        '.. glossary::\n'
        '\n'
        '   term1\n'
        '       description\n'
        '   .. term2\n'
        '       description\n'
        '       description\n'
    )
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            [
                glossary,
                definition_list,
                definition_list_item,
                ([term, ('term1', index)], definition),
            ],
        ),
    )
    assert_node(doctree[0][0][0][1], [nodes.definition, nodes.paragraph, 'description'])


@pytest.mark.sphinx('html', testroot='_blank')
def test_glossary_comment2(app):
    text = (
        '.. glossary::\n'
        '\n'
        '   term1\n'
        '       description\n'
        '\n'
        '   .. term2\n'
        '   term3\n'
        '       description\n'
        '       description\n'
    )
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            [
                glossary,
                definition_list,
                (
                    [definition_list_item, ([term, ('term1', index)], definition)],
                    [definition_list_item, ([term, ('term3', index)], definition)],
                ),
            ],
        ),
    )
    assert_node(doctree[0][0][0][1], [nodes.definition, nodes.paragraph, 'description'])
    assert_node(
        doctree[0][0][1][1],
        [nodes.definition, nodes.paragraph, 'description\ndescription'],
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_glossary_sorted(app):
    text = (
        '.. glossary::\n'
        '   :sorted:\n'
        '\n'
        '   term3\n'
        '       description\n'
        '\n'
        '   term2\n'
        '   term1\n'
        '       description\n'
    )
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            [
                glossary,
                definition_list,
                (
                    [
                        definition_list_item,
                        (
                            [term, ('term2', index)],
                            [term, ('term1', index)],
                            definition,
                        ),
                    ],
                    [definition_list_item, ([term, ('term3', index)], definition)],
                ),
            ],
        ),
    )
    assert_node(doctree[0][0][0][2], [nodes.definition, nodes.paragraph, 'description'])
    assert_node(doctree[0][0][1][1], [nodes.definition, nodes.paragraph, 'description'])


@pytest.mark.sphinx('html', testroot='_blank')
def test_glossary_alphanumeric(app: SphinxTestApp) -> None:
    text = '.. glossary::\n\n   1\n   /\n'
    restructuredtext.parse(app, text)
    objects = list(app.env.domains.standard_domain.get_objects())
    assert ('1', '1', 'term', 'index', 'term-1', -1) in objects
    assert ('/', '/', 'term', 'index', 'term-0', -1) in objects


@pytest.mark.sphinx('html', testroot='_blank')
def test_glossary_conflicted_labels(app: SphinxTestApp) -> None:
    text = '.. _term-foo:\n.. glossary::\n\n   foo\n'
    restructuredtext.parse(app, text)
    objects = list(app.env.domains.standard_domain.get_objects())
    assert ('foo', 'foo', 'term', 'index', 'term-0', -1) in objects


@pytest.mark.sphinx('html', testroot='_blank')
def test_cmdoption(app: SphinxTestApp) -> None:
    text = '.. program:: ls\n\n.. option:: -l\n'
    domain = app.env.domains.standard_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [desc_signature, ([desc_name, '-l'], [desc_addname, ()])],
                    [desc_content, ()],
                ),
            ],
        ),
    )
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('pair', 'ls command line option; -l', 'cmdoption-ls-l', '', None)],
    )
    assert ('ls', '-l') in domain.progoptions
    assert domain.progoptions['ls', '-l'] == ('index', 'cmdoption-ls-l')


@pytest.mark.sphinx('html', testroot='_blank')
def test_cmdoption_for_None(app: SphinxTestApp) -> None:
    text = '.. program:: ls\n.. program:: None\n\n.. option:: -l\n'
    domain = app.env.domains.standard_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [desc_signature, ([desc_name, '-l'], [desc_addname, ()])],
                    [desc_content, ()],
                ),
            ],
        ),
    )
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('pair', 'command line option; -l', 'cmdoption-l', '', None)],
    )
    assert (None, '-l') in domain.progoptions
    assert domain.progoptions[None, '-l'] == ('index', 'cmdoption-l')


@pytest.mark.sphinx('html', testroot='_blank')
def test_multiple_cmdoptions(app: SphinxTestApp) -> None:
    text = '.. program:: cmd\n\n.. option:: -o directory, --output directory\n'
    domain = app.env.domains.standard_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, '-o'],
                            [desc_addname, ' directory'],
                            [desc_addname, ', '],
                            [desc_name, '--output'],
                            [desc_addname, ' directory'],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[
            ('pair', 'cmd command line option; -o', 'cmdoption-cmd-o', '', None),
            ('pair', 'cmd command line option; --output', 'cmdoption-cmd-o', '', None),
        ],
    )
    assert ('cmd', '-o') in domain.progoptions
    assert ('cmd', '--output') in domain.progoptions
    assert domain.progoptions['cmd', '-o'] == ('index', 'cmdoption-cmd-o')
    assert domain.progoptions['cmd', '--output'] == ('index', 'cmdoption-cmd-o')


@pytest.mark.sphinx('html', testroot='_blank')
def test_disabled_docref(app: SphinxTestApp) -> None:
    text = ':doc:`index`\n:doc:`!index`\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            [
                nodes.paragraph,
                ([pending_xref, nodes.inline, 'index'], '\n', [nodes.inline, 'index']),
            ],
        ),
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_labeled_rubric(app: SphinxTestApp) -> None:
    text = '.. _label:\n.. rubric:: blah *blah* blah\n'
    restructuredtext.parse(app, text)

    domain = app.env.domains.standard_domain
    assert 'label' in domain.labels
    assert domain.labels['label'] == ('index', 'label', 'blah blah blah')


@pytest.mark.sphinx('html', testroot='_blank')
def test_labeled_definition(app: SphinxTestApp) -> None:
    text = (
        '.. _label1:\n'
        '\n'
        'Foo blah *blah* blah\n'
        '  Definition\n'
        '\n'
        '.. _label2:\n'
        '\n'
        'Bar blah *blah* blah\n'
        '  Definition\n'
        '\n'
    )
    restructuredtext.parse(app, text)

    domain = app.env.domains.standard_domain
    assert 'label1' in domain.labels
    assert domain.labels['label1'] == ('index', 'label1', 'Foo blah blah blah')
    assert 'label2' in domain.labels
    assert domain.labels['label2'] == ('index', 'label2', 'Bar blah blah blah')


@pytest.mark.sphinx('html', testroot='_blank')
def test_labeled_field(app: SphinxTestApp) -> None:
    text = (
        '.. _label1:\n'
        '\n'
        ':Foo blah *blah* blah:\n'
        '  Definition\n'
        '\n'
        '.. _label2:\n'
        '\n'
        ':Bar blah *blah* blah:\n'
        '  Definition\n'
        '\n'
    )
    restructuredtext.parse(app, text)

    domain = app.env.domains.standard_domain
    assert 'label1' in domain.labels
    assert domain.labels['label1'] == ('index', 'label1', 'Foo blah blah blah')
    assert 'label2' in domain.labels
    assert domain.labels['label2'] == ('index', 'label2', 'Bar blah blah blah')


@pytest.mark.sphinx(
    'html',
    testroot='manpage_url',
    confoverrides={'manpages_url': 'https://example.com/{page}.{section}'},
)
def test_html_manpage(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert (
        '<em class="manpage">'
        '<a class="manpage reference external" href="https://example.com/man.1">man(1)</a>'
        '</em>'
    ) in content
    assert (
        '<em class="manpage">'
        '<a class="manpage reference external" href="https://example.com/ls.1">ls.1</a>'
        '</em>'
    ) in content
    assert (
        '<em class="manpage">'
        '<a class="manpage reference external" href="https://example.com/sphinx.">sphinx</a>'
        '</em>'
    ) in content
    assert (
        '<em class="manpage">'
        '<a class="manpage reference external" href="https://example.com/bsd-mailx/mailx.1">mailx(1)</a>'
        '</em>'
    ) in content
    assert '<em class="manpage">man(1)</em>' in content



"""Test sphinx.ext.inheritance_diagram extension."""

from __future__ import annotations

import re
import sys
import zlib
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from sphinx.ext.inheritance_diagram import (
    InheritanceDiagram,
    InheritanceException,
    import_classes,
)
from sphinx.ext.intersphinx._load import load_mappings, validate_intersphinx_mapping

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='inheritance')
@pytest.mark.usefixtures('if_graphviz_found')
def test_inheritance_diagram(app: SphinxTestApp) -> None:
    # monkey-patch InheritaceDiagram.run() so we can get access to its
    # results.
    orig_run = InheritanceDiagram.run
    graphs = {}

    def new_run(self):
        result = orig_run(self)
        node = result[0]
        source = Path(node.document.current_source).stem
        graphs[source] = node['graph']
        return result

    InheritanceDiagram.run = new_run  # type: ignore[method-assign]

    try:
        app.build(force_all=True)
    finally:
        InheritanceDiagram.run = orig_run  # type: ignore[method-assign]

    assert app.statuscode == 0

    html_warnings = app.warning.getvalue()
    assert html_warnings == ''

    # note: it is better to split these asserts into separate test functions
    # but I can't figure out how to build only a specific .rst file

    # basic inheritance diagram showing all classes
    for cls in graphs['basic_diagram'].class_info:
        # use in b/c traversing order is different sometimes
        assert cls in {
            ('dummy.test.A', 'dummy.test.A', (), None),
            ('dummy.test.F', 'dummy.test.F', ('dummy.test.C',), None),
            ('dummy.test.C', 'dummy.test.C', ('dummy.test.A',), None),
            ('dummy.test.E', 'dummy.test.E', ('dummy.test.B',), None),
            ('dummy.test.D', 'dummy.test.D', ('dummy.test.B', 'dummy.test.C'), None),
            ('dummy.test.B', 'dummy.test.B', ('dummy.test.A',), None),
        }

    # inheritance diagram using :parts: 1 option
    for cls in graphs['diagram_w_parts'].class_info:
        assert cls in {
            ('A', 'dummy.test.A', (), None),
            ('F', 'dummy.test.F', ('C',), None),
            ('C', 'dummy.test.C', ('A',), None),
            ('E', 'dummy.test.E', ('B',), None),
            ('D', 'dummy.test.D', ('B', 'C'), None),
            ('B', 'dummy.test.B', ('A',), None),
        }

    # inheritance diagram with 1 top class
    # :top-classes: dummy.test.B
    # rendering should be
    #       A
    #        \
    #     B   C
    #    / \ / \
    #   E   D   F
    #
    for cls in graphs['diagram_w_1_top_class'].class_info:
        assert cls in {
            ('dummy.test.A', 'dummy.test.A', (), None),
            ('dummy.test.F', 'dummy.test.F', ('dummy.test.C',), None),
            ('dummy.test.C', 'dummy.test.C', ('dummy.test.A',), None),
            ('dummy.test.E', 'dummy.test.E', ('dummy.test.B',), None),
            ('dummy.test.D', 'dummy.test.D', ('dummy.test.B', 'dummy.test.C'), None),
            ('dummy.test.B', 'dummy.test.B', (), None),
        }

    # inheritance diagram with 2 top classes
    # :top-classes: dummy.test.B, dummy.test.C
    # Note: we're specifying separate classes, not the entire module here
    # rendering should be
    #
    #     B   C
    #    / \ / \
    #   E   D   F
    #
    for cls in graphs['diagram_w_2_top_classes'].class_info:
        assert cls in {
            ('dummy.test.F', 'dummy.test.F', ('dummy.test.C',), None),
            ('dummy.test.C', 'dummy.test.C', (), None),
            ('dummy.test.E', 'dummy.test.E', ('dummy.test.B',), None),
            ('dummy.test.D', 'dummy.test.D', ('dummy.test.B', 'dummy.test.C'), None),
            ('dummy.test.B', 'dummy.test.B', (), None),
        }

    # inheritance diagram with 2 top classes and specifying the entire module
    # rendering should be
    #
    #       A
    #     B   C
    #    / \ / \
    #   E   D   F
    #
    # Note: dummy.test.A is included in the graph before its descendants are even processed
    # b/c we've specified to load the entire module. The way InheritanceGraph works it is very
    # hard to exclude parent classes once after they have been included in the graph.
    # If you'd like to not show class A in the graph don't specify the entire module.
    # this is a known issue.
    for cls in graphs['diagram_module_w_2_top_classes'].class_info:
        assert cls in {
            ('dummy.test.F', 'dummy.test.F', ('dummy.test.C',), None),
            ('dummy.test.C', 'dummy.test.C', (), None),
            ('dummy.test.E', 'dummy.test.E', ('dummy.test.B',), None),
            ('dummy.test.D', 'dummy.test.D', ('dummy.test.B', 'dummy.test.C'), None),
            ('dummy.test.B', 'dummy.test.B', (), None),
            ('dummy.test.A', 'dummy.test.A', (), None),
        }

    # inheritance diagram involving a base class nested within another class
    for cls in graphs['diagram_w_nested_classes'].class_info:
        assert cls in {
            ('dummy.test_nested.A', 'dummy.test_nested.A', (), None),
            (
                'dummy.test_nested.C',
                'dummy.test_nested.C',
                ('dummy.test_nested.A.B',),
                None,
            ),
            ('dummy.test_nested.A.B', 'dummy.test_nested.A.B', (), None),
        }


# An external inventory to test intersphinx links in inheritance diagrams
external_inventory = b"""\
# Sphinx inventory version 2
# Project: external
# Version: 1.0
# The remainder of this file is compressed using zlib.
""" + zlib.compress(b"""\
external.other.Bob py:class 1 foo.html#external.other.Bob -
""")


@pytest.mark.sphinx('html', testroot='ext-inheritance_diagram')
@pytest.mark.usefixtures('if_graphviz_found')
def test_inheritance_diagram_png_html(tmp_path, app):
    inv_file = tmp_path / 'inventory'
    inv_file.write_bytes(external_inventory)
    app.config.intersphinx_mapping = {
        'example': ('https://example.org', str(inv_file)),
    }
    app.config.intersphinx_cache_limit = 0
    validate_intersphinx_mapping(app, app.config)
    load_mappings(app)

    app.build(force_all=True)

    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    base_maps = re.findall('<map .+\n.+\n</map>', content)

    pattern = (
        '<figure class="align-default" id="id1">\n'
        '<div class="graphviz">'
        '<img src="_images/inheritance-\\w+.png" alt="Inheritance diagram of test.Foo" '
        'class="inheritance graphviz" /></div>\n<figcaption>\n<p>'
        '<span class="caption-text">Test Foo!</span><a class="headerlink" href="#id1" '
        'title="Link to this image">\xb6</a></p>\n</figcaption>\n</figure>\n'
    )
    assert re.search(pattern, content, re.MULTILINE)

    subdir_content = (app.outdir / 'subdir/page1.html').read_text(encoding='utf8')
    subdir_maps = re.findall('<map .+\n.+\n</map>', subdir_content)
    subdir_maps = [
        re.sub('href="(\\S+)"', 'href="subdir/\\g<1>"', s) for s in subdir_maps
    ]

    # Go through the clickmap for every PNG inheritance diagram
    for diagram_content in base_maps + subdir_maps:
        # Verify that an intersphinx link was created via the external inventory
        if 'subdir.' in diagram_content:
            assert 'https://example.org' in diagram_content

        # Extract every link in the inheritance diagram
        for href in re.findall('href="(\\S+?)"', diagram_content):
            if '://' in href:
                # Verify that absolute URLs are not prefixed with ../
                assert href.startswith('https://example.org/')
            else:
                # Verify that relative URLs point to existing documents
                reluri = href.rsplit('#', 1)[0]  # strip the anchor at the end
                assert (app.outdir / reluri).exists()


@pytest.mark.sphinx(
    'html',
    testroot='ext-inheritance_diagram',
    confoverrides={'graphviz_output_format': 'svg'},
)
@pytest.mark.usefixtures('if_graphviz_found')
def test_inheritance_diagram_svg_html(tmp_path, app):
    inv_file = tmp_path / 'inventory'
    inv_file.write_bytes(external_inventory)
    app.config.intersphinx_mapping = {
        'subdir': ('https://example.org', str(inv_file)),
    }
    app.config.intersphinx_cache_limit = 0
    validate_intersphinx_mapping(app, app.config)
    load_mappings(app)

    app.build(force_all=True)

    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    base_svgs = re.findall('<object data="(_images/inheritance-\\w+.svg?)"', content)

    pattern = (
        '<figure class="align-default" id="id1">\n'
        '<div class="graphviz">'
        '<object data="_images/inheritance-\\w+.svg" '
        'type="image/svg\\+xml" class="inheritance graphviz">\n'
        '<p class="warning">Inheritance diagram of test.Foo</p>'
        '</object></div>\n<figcaption>\n<p><span class="caption-text">'
        'Test Foo!</span><a class="headerlink" href="#id1" '
        'title="Link to this image">\xb6</a></p>\n</figcaption>\n</figure>\n'
    )

    assert re.search(pattern, content, re.MULTILINE)

    subdir_content = (app.outdir / 'subdir/page1.html').read_text(encoding='utf8')
    subdir_svgs = re.findall(
        '<object data="../(_images/inheritance-\\w+.svg?)"', subdir_content
    )

    # Go through every SVG inheritance diagram
    for diagram in base_svgs + subdir_svgs:
        diagram_content = (app.outdir / diagram).read_text(encoding='utf8')

        # Verify that an intersphinx link was created via the external inventory
        if 'subdir.' in diagram_content:
            assert 'https://example.org' in diagram_content

        # Extract every link in the inheritance diagram
        for href in re.findall('href="(\\S+?)"', diagram_content):
            if '://' in href:
                # Verify that absolute URLs are not prefixed with ../
                assert href.startswith('https://example.org/')
            else:
                # Verify that relative URLs point to existing documents
                reluri = href.rsplit('#', 1)[0]  # strip the anchor at the end
                abs_uri = (app.outdir / app.builder.imagedir / reluri).resolve()
                assert abs_uri.exists()


@pytest.mark.sphinx('latex', testroot='ext-inheritance_diagram')
@pytest.mark.usefixtures('if_graphviz_found')
def test_inheritance_diagram_latex(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    content = (app.outdir / 'projectnamenotset.tex').read_text(encoding='utf8')

    pattern = (
        '\\\\begin{figure}\\[htbp]\n\\\\centering\n\\\\capstart\n\n'
        '\\\\sphinxincludegraphics\\[\\]{inheritance-\\w+.pdf}\n'
        '\\\\caption{Test Foo!}\\\\label{\\\\detokenize{index:id1}}\\\\end{figure}'
    )
    assert re.search(pattern, content, re.MULTILINE)


@pytest.mark.sphinx(
    'html',
    testroot='ext-inheritance_diagram',
    srcdir='ext-inheritance_diagram-alias',
)
@pytest.mark.usefixtures('if_graphviz_found')
def test_inheritance_diagram_latex_alias(app):
    app.config.inheritance_alias = {'test.Foo': 'alias.Foo'}
    app.build(force_all=True)

    doc = app.env.get_and_resolve_doctree('index', app)
    aliased_graph = doc.children[0].children[3]['graph'].class_info
    assert len(aliased_graph) == 4
    assert (
        'test.DocSubDir2',
        'test.DocSubDir2',
        ('test.DocSubDir1',),
        None,
    ) in aliased_graph
    assert (
        'test.DocSubDir1',
        'test.DocSubDir1',
        ('test.DocHere',),
        None,
    ) in aliased_graph
    assert ('test.DocHere', 'test.DocHere', ('alias.Foo',), None) in aliased_graph
    assert ('alias.Foo', 'alias.Foo', (), None) in aliased_graph

    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    pattern = (
        '<figure class="align-default" id="id1">\n'
        '<div class="graphviz">'
        '<img src="_images/inheritance-\\w+.png" alt="Inheritance diagram of test.Foo" '
        'class="inheritance graphviz" /></div>\n<figcaption>\n<p>'
        '<span class="caption-text">Test Foo!</span><a class="headerlink" href="#id1" '
        'title="Link to this image">\xb6</a></p>\n</figcaption>\n</figure>\n'
    )
    assert re.search(pattern, content, re.MULTILINE)


def test_import_classes(rootdir):
    from sphinx.parsers import Parser, RSTParser
    from sphinx.util.i18n import CatalogInfo

    saved_path = sys.path.copy()
    sys.path.insert(0, str(rootdir / 'test-ext-inheritance_diagram'))
    try:
        from example.sphinx import DummyClass  # type: ignore[import-not-found]

        # got exception for unknown class or module
        with pytest.raises(InheritanceException):
            import_classes('unknown', None)
        with pytest.raises(InheritanceException):
            import_classes('unknown.Unknown', None)

        # got exception InheritanceException for wrong class or module
        # not AttributeError
        # See: https://github.com/sphinx-doc/sphinx/issues/4019
        with pytest.raises(InheritanceException):
            import_classes('unknown', '.')
        with pytest.raises(InheritanceException):
            import_classes('unknown.Unknown', '.')
        with pytest.raises(InheritanceException):
            import_classes('.', None)

        # a module having no classes
        classes = import_classes('sphinx', None)
        assert classes == []

        classes = import_classes('sphinx', 'foo')
        assert classes == []

        # all of classes in the module
        classes = import_classes('sphinx.parsers', None)
        assert set(classes) == {Parser, RSTParser}

        # specified class in the module
        classes = import_classes('sphinx.parsers.Parser', None)
        assert classes == [Parser]

        # specified class in current module
        classes = import_classes('Parser', 'sphinx.parsers')
        assert classes == [Parser]

        # relative module name to current module
        classes = import_classes('i18n.CatalogInfo', 'sphinx.util')
        assert classes == [CatalogInfo]

        # got exception for functions
        with pytest.raises(InheritanceException):
            import_classes('encode_uri', 'sphinx.util')

        # import submodule on current module
        # See: https://github.com/sphinx-doc/sphinx/issues/3164
        classes = import_classes('sphinx', 'example')
        assert classes == [DummyClass]
    finally:
        sys.path[:] = saved_path



"""Test the HTML builder and check output against XPath."""

from __future__ import annotations

import re
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

import sphinx.builders.html
from sphinx.builders.html._assets import _file_checksum
from sphinx.errors import ThemeError

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='html_assets')
def test_html_assets(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    # exclude_path and its family
    assert not (app.outdir / 'static' / 'index.html').exists()
    assert not (app.outdir / 'extra' / 'index.html').exists()

    # html_static_path
    assert not (app.outdir / '_static' / '.htaccess').exists()
    assert not (app.outdir / '_static' / '.htpasswd').exists()
    assert (app.outdir / '_static' / 'API.html').exists()
    assert (app.outdir / '_static' / 'API.html').read_text(
        encoding='utf8'
    ) == 'Sphinx-1.4.4'
    assert (app.outdir / '_static' / 'css' / 'style.css').exists()
    assert (app.outdir / '_static' / 'js' / 'custom.js').exists()
    assert (app.outdir / '_static' / 'rimg.png').exists()
    assert not (app.outdir / '_static' / '_build' / 'index.html').exists()
    assert (app.outdir / '_static' / 'background.png').exists()
    assert not (app.outdir / '_static' / 'subdir' / '.htaccess').exists()
    assert not (app.outdir / '_static' / 'subdir' / '.htpasswd').exists()

    # html_extra_path
    assert (app.outdir / '.htaccess').exists()
    assert not (app.outdir / '.htpasswd').exists()
    assert (app.outdir / 'API.html.jinja').exists()
    assert (app.outdir / 'css/style.css').exists()
    assert (app.outdir / 'rimg.png').exists()
    assert not (app.outdir / '_build' / 'index.html').exists()
    assert (app.outdir / 'background.png').exists()
    assert (app.outdir / 'subdir' / '.htaccess').exists()
    assert not (app.outdir / 'subdir' / '.htpasswd').exists()

    # html_css_files
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert (
        '<link rel="stylesheet" type="text/css" href="_static/css/style.css" />'
    ) in content
    assert (
        '<link media="print" rel="stylesheet" title="title" type="text/css" '
        'href="https://example.com/custom.css" />'
    ) in content

    # html_js_files
    assert '<script src="_static/js/custom.js"></script>' in content
    assert (
        '<script async="async" src="https://example.com/script.js"></script>'
    ) in content


@pytest.mark.sphinx('html', testroot='html_assets')
def test_assets_order(app, monkeypatch):
    monkeypatch.setattr(sphinx.builders.html, '_file_checksum', lambda o, f: '')

    app.add_css_file('normal.css')
    app.add_css_file('early.css', priority=100)
    app.add_css_file('late.css', priority=750)
    app.add_css_file('lazy.css', priority=900)
    app.add_js_file('normal.js')
    app.add_js_file('early.js', priority=100)
    app.add_js_file('late.js', priority=750)
    app.add_js_file('lazy.js', priority=900)

    app.build(force_all=True)
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    # css_files
    expected = [
        '_static/early.css',
        '_static/pygments.css',
        'https://example.com/custom.css',
        '_static/normal.css',
        '_static/late.css',
        '_static/css/style.css',
        '_static/lazy.css',
    ]
    pattern = '.*'.join(f'href="{re.escape(f)}"' for f in expected)
    assert re.search(pattern, content, re.DOTALL), content

    # js_files
    expected = [
        '_static/early.js',
        '_static/doctools.js',
        '_static/sphinx_highlight.js',
        'https://example.com/script.js',
        '_static/normal.js',
        '_static/late.js',
        '_static/js/custom.js',
        '_static/lazy.js',
    ]
    pattern = '.*'.join(f'src="{re.escape(f)}"' for f in expected)
    assert re.search(pattern, content, re.DOTALL), content


@pytest.mark.sphinx('html', testroot='html_file_checksum')
def test_file_checksum(app: SphinxTestApp) -> None:
    app.add_css_file('stylesheet-a.css')
    app.add_css_file('stylesheet-b.css')
    app.add_css_file('https://example.com/custom.css')
    app.add_js_file('script.js')
    app.add_js_file('empty.js')
    app.add_js_file('https://example.com/script.js')

    app.build(force_all=True)
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    # checksum for local files
    assert (
        '<link rel="stylesheet" type="text/css" href="_static/stylesheet-a.css?v=e575b6df" />'
    ) in content
    assert (
        '<link rel="stylesheet" type="text/css" href="_static/stylesheet-b.css?v=a2d5cc0f" />'
    ) in content
    assert '<script src="_static/script.js?v=48278d48"></script>' in content

    # empty files have no checksum
    assert '<script src="_static/empty.js"></script>' in content

    # no checksum for hyperlinks
    assert (
        '<link rel="stylesheet" type="text/css" href="https://example.com/custom.css" />'
    ) in content
    assert '<script src="https://example.com/script.js"></script>' in content


def test_file_checksum_query_string() -> None:
    with pytest.raises(
        ThemeError,
        match='Local asset file paths must not contain query strings',
    ):
        _file_checksum(Path(), 'with_query_string.css?dead_parrots=1')

    with pytest.raises(
        ThemeError,
        match='Local asset file paths must not contain query strings',
    ):
        _file_checksum(Path(), 'with_query_string.js?dead_parrots=1')

    with pytest.raises(
        ThemeError,
        match='Local asset file paths must not contain query strings',
    ):
        _file_checksum(Path.cwd(), '_static/with_query_string.css?dead_parrots=1')

    with pytest.raises(
        ThemeError,
        match='Local asset file paths must not contain query strings',
    ):
        _file_checksum(Path.cwd(), '_static/with_query_string.js?dead_parrots=1')


@pytest.mark.sphinx('html', testroot='html_assets')
def test_javscript_loading_method(app: SphinxTestApp) -> None:
    app.add_js_file('normal.js')
    app.add_js_file('early.js', loading_method='async')
    app.add_js_file('late.js', loading_method='defer')

    app.build(force_all=True)
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    assert '<script src="_static/normal.js"></script>' in content
    assert '<script async="async" src="_static/early.js"></script>' in content
    assert '<script defer="defer" src="_static/late.js"></script>' in content



"""A module with a trailing underscore"""


class SomeClass_:
    """A class with a trailing underscore"""


def some_function_(some_arg_):
    """A function with a trailing underscore in name and argument"""



"""Test the HTML builder and check output against XPath."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from sphinx.builders.html import StandaloneHTMLBuilder

from tests.test_builders.xpath_html_util import _intradocument_hyperlink_check
from tests.test_builders.xpath_util import check_xpath

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='toctree-glob')
def test_relations(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    assert app.builder.relations['index'] == [None, None, 'foo']
    assert app.builder.relations['foo'] == ['index', 'index', 'bar/index']
    assert app.builder.relations['bar/index'] == ['index', 'foo', 'bar/bar_1']
    assert app.builder.relations['bar/bar_1'] == ['bar/index', 'bar/index', 'bar/bar_2']
    assert app.builder.relations['bar/bar_2'] == ['bar/index', 'bar/bar_1', 'bar/bar_3']
    assert app.builder.relations['bar/bar_3'] == [
        'bar/index',
        'bar/bar_2',
        'bar/bar_4/index',
    ]
    assert app.builder.relations['bar/bar_4/index'] == ['bar/index', 'bar/bar_3', 'baz']
    assert app.builder.relations['baz'] == ['index', 'bar/bar_4/index', 'qux/index']
    assert app.builder.relations['qux/index'] == ['index', 'baz', 'qux/qux_1']
    assert app.builder.relations['qux/qux_1'] == ['qux/index', 'qux/index', 'qux/qux_2']
    assert app.builder.relations['qux/qux_2'] == ['qux/index', 'qux/qux_1', None]
    assert 'quux' not in app.builder.relations


@pytest.mark.sphinx('singlehtml', testroot='toctree-empty')
def test_singlehtml_toctree(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    try:
        app.builder._get_local_toctree('index')
    except AttributeError:
        pytest.fail('Unexpected AttributeError in app.builder.fix_refuris')


@pytest.mark.sphinx(
    'html',
    testroot='toctree',
    srcdir='numbered-toctree',
)
def test_numbered_toctree(app: SphinxTestApp) -> None:
    # give argument to :numbered: option
    index = (app.srcdir / 'index.rst').read_text(encoding='utf8')
    index = re.sub(':numbered:.*', ':numbered: 1', index)
    (app.srcdir / 'index.rst').write_text(index, encoding='utf8')
    app.build(force_all=True)


@pytest.mark.parametrize(
    'expect',
    [
        # internal references should be same-document; external should not
        (".//a[@class='reference internal']", _intradocument_hyperlink_check),
        (".//a[@class='reference external']", r'https?://'),
    ],
)
@pytest.mark.sphinx('singlehtml', testroot='toctree')
def test_singlehtml_hyperlinks(app, cached_etree_parse, expect):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / 'index.html'), 'index.html', *expect)


@pytest.mark.sphinx(
    'html',
    testroot='toctree-multiple-parents',
    confoverrides={'html_theme': 'alabaster'},
)
def test_toctree_multiple_parents(app, cached_etree_parse):
    # The lexicographically greatest parent of the document in global toctree
    # should be chosen, regardless of the order in which files are read
    with patch.object(app.builder, '_read_serial') as m:
        # Read files in reversed order
        _read_serial = type(app.builder)._read_serial
        m.side_effect = lambda docnames: _read_serial(app.builder, docnames[::-1])
        app.build()
        # Check if charlie is a child of delta in charlie.html
        xpath_delta_children = (
            ".//ul[@class='current']//a[@href='delta.html']/../ul/li//a"
        )
        etree = cached_etree_parse(app.outdir / 'charlie.html')
        check_xpath(etree, 'charlie.html', xpath=xpath_delta_children, check='Charlie')



"""Test sphinx.extension module."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sphinx.errors import VersionRequirementError
from sphinx.extension import Extension, verify_needs_extensions

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='root')
def test_needs_extensions(app: SphinxTestApp) -> None:
    # empty needs_extensions
    assert app.config.needs_extensions == {}
    verify_needs_extensions(app, app.config)

    # needs_extensions fulfilled
    app.config.needs_extensions = {'test.extension': '3.9'}
    app.extensions['test.extension'] = Extension(
        'test.extension', 'test.extension', version='3.10'
    )
    verify_needs_extensions(app, app.config)

    # needs_extensions not fulfilled
    app.config.needs_extensions = {'test.extension': '3.11'}
    app.extensions['test.extension'] = Extension(
        'test.extension', 'test.extension', version='3.10'
    )
    with pytest.raises(VersionRequirementError):
        verify_needs_extensions(app, app.config)



"""Tests util.typing functions."""

from __future__ import annotations

import ctypes
import dataclasses
import sys
import typing as t
import zipfile
from collections import abc
from contextvars import Context, ContextVar, Token
from enum import Enum
from io import (
    BufferedRandom,
    BufferedReader,
    BufferedRWPair,
    BufferedWriter,
    BytesIO,
    FileIO,
    StringIO,
    TextIOWrapper,
)
from json import JSONDecoder, JSONEncoder
from lzma import LZMACompressor, LZMADecompressor
from multiprocessing import Process
from numbers import Integral
from pathlib import (
    Path,
    PosixPath,
    PurePath,
    PurePosixPath,
    PureWindowsPath,
    WindowsPath,
)
from pickle import Pickler, Unpickler
from struct import Struct
from types import (
    AsyncGeneratorType,
    BuiltinFunctionType,
    BuiltinMethodType,
    CellType,
    ClassMethodDescriptorType,
    CodeType,
    CoroutineType,
    EllipsisType,
    FrameType,
    FunctionType,
    GeneratorType,
    GetSetDescriptorType,
    LambdaType,
    MappingProxyType,
    MemberDescriptorType,
    MethodDescriptorType,
    MethodType,
    MethodWrapperType,
    ModuleType,
    NoneType,
    NotImplementedType,
    TracebackType,
    WrapperDescriptorType,
)
from typing import (
    Annotated,
    Any,
    Dict,
    ForwardRef,
    List,
    Literal,
    NewType,
    Optional,
    ParamSpec,
    Tuple,
    TypeVar,
    Union,
)
from weakref import WeakSet

from sphinx.ext.autodoc.mock import mock
from sphinx.util.typing import _INVALID_BUILTIN_CLASSES, restify, stringify_annotation


class MyClass1:
    pass


class MyClass2(MyClass1):
    __qualname__ = '<MyClass2>'


class MyEnum(Enum):
    a = 1


T = TypeVar('T')
MyInt = NewType('MyInt', int)


class MyList(List[T]):
    pass


class BrokenType:
    __args__ = int


@dataclasses.dataclass(frozen=True)
class Gt:
    gt: float


def test_restify() -> None:
    assert restify(int) == ':py:class:`int`'
    assert restify(int, 'smart') == ':py:class:`int`'

    assert restify(str) == ':py:class:`str`'
    assert restify(str, 'smart') == ':py:class:`str`'

    assert restify(None) == ':py:obj:`None`'
    assert restify(None, 'smart') == ':py:obj:`None`'

    assert restify(Integral) == ':py:class:`numbers.Integral`'
    assert restify(Integral, 'smart') == ':py:class:`~numbers.Integral`'

    assert restify(Struct) == ':py:class:`struct.Struct`'
    assert restify(Struct, 'smart') == ':py:class:`~struct.Struct`'

    assert restify(TracebackType) == ':py:class:`types.TracebackType`'
    assert restify(TracebackType, 'smart') == ':py:class:`~types.TracebackType`'

    assert restify(Path) == ':py:class:`pathlib.Path`'
    assert restify(Path, 'smart') == ':py:class:`~pathlib.Path`'

    assert restify(Any) == ':py:obj:`~typing.Any`'
    assert restify(Any, 'smart') == ':py:obj:`~typing.Any`'

    assert restify('str') == 'str'
    assert restify('str', 'smart') == 'str'


def test_is_invalid_builtin_class() -> None:
    # if these tests start failing, it means that the __module__
    # of one of these classes has changed, and _INVALID_BUILTIN_CLASSES
    # in sphinx.util.typing needs to be updated.
    invalid_types = (
        # contextvars
        Context,
        ContextVar,
        Token,
        # ctypes
        ctypes.Array,
        ctypes.Structure,
        ctypes.Union,
        # io
        FileIO,
        BytesIO,
        StringIO,
        BufferedReader,
        BufferedWriter,
        BufferedRWPair,
        BufferedRandom,
        TextIOWrapper,
        # json
        JSONDecoder,
        JSONEncoder,
        # lzma
        LZMACompressor,
        LZMADecompressor,
        # multiprocessing
        Process,
        # pickle
        Pickler,
        Unpickler,
        # struct
        Struct,
        # types
        AsyncGeneratorType,
        BuiltinFunctionType,
        BuiltinMethodType,
        CellType,
        ClassMethodDescriptorType,
        CodeType,
        CoroutineType,
        EllipsisType,
        FrameType,
        FunctionType,
        GeneratorType,
        GetSetDescriptorType,
        LambdaType,
        MappingProxyType,
        MemberDescriptorType,
        MethodDescriptorType,
        MethodType,
        MethodWrapperType,
        ModuleType,
        NoneType,
        NotImplementedType,
        TracebackType,
        WrapperDescriptorType,
        # weakref
        WeakSet,
    )
    if sys.version_info[:2] >= (3, 12):
        invalid_types += (
            # zipfile
            zipfile.Path,
            zipfile.CompleteDirs,
        )
    if sys.version_info[:2] >= (3, 13):
        invalid_types += (
            # pathlib
            Path,
            PosixPath,
            PurePath,
            PurePosixPath,
            PureWindowsPath,
            WindowsPath,
        )

    invalid_names = {(cls.__module__, cls.__qualname__) for cls in invalid_types}
    if sys.version_info[:2] < (3, 13):
        invalid_names |= {
            ('pathlib._local', 'Path'),
            ('pathlib._local', 'PosixPath'),
            ('pathlib._local', 'PurePath'),
            ('pathlib._local', 'PurePosixPath'),
            ('pathlib._local', 'PureWindowsPath'),
            ('pathlib._local', 'WindowsPath'),
        }
    if sys.version_info[:2] < (3, 12):
        invalid_names |= {
            ('zipfile._path', 'Path'),
            ('zipfile._path', 'CompleteDirs'),
        }
    assert _INVALID_BUILTIN_CLASSES.keys() == invalid_names


def test_restify_type_hints_containers():
    assert restify(List) == ':py:class:`~typing.List`'
    assert restify(Dict) == ':py:class:`~typing.Dict`'
    assert restify(List[int]) == ':py:class:`~typing.List`\\ [:py:class:`int`]'
    assert restify(List[str]) == ':py:class:`~typing.List`\\ [:py:class:`str`]'
    assert restify(Dict[str, float]) == (
        ':py:class:`~typing.Dict`\\ [:py:class:`str`, :py:class:`float`]'
    )
    assert restify(Tuple[str, str, str]) == (
        ':py:class:`~typing.Tuple`\\ '
        '[:py:class:`str`, :py:class:`str`, '
        ':py:class:`str`]'
    )
    ann_rst = restify(Tuple[str, ...])
    assert ann_rst == ':py:class:`~typing.Tuple`\\ [:py:class:`str`, ...]'

    assert restify(Tuple[()]) == ':py:class:`~typing.Tuple`'

    assert restify(List[Dict[str, Tuple]]) == (
        ':py:class:`~typing.List`\\ '
        '[:py:class:`~typing.Dict`\\ '
        '[:py:class:`str`, :py:class:`~typing.Tuple`]]'
    )
    assert restify(MyList[Tuple[int, int]]) == (
        ':py:class:`tests.test_util.test_util_typing.MyList`\\ '
        '[:py:class:`~typing.Tuple`\\ '
        '[:py:class:`int`, :py:class:`int`]]'
    )
    assert restify(t.Generator[None, None, None]) == (
        ':py:class:`~typing.Generator`\\ '
        '[:py:obj:`None`, :py:obj:`None`, '
        ':py:obj:`None`]'
    )
    assert restify(abc.Generator[None, None, None]) == (
        ':py:class:`collections.abc.Generator`\\ '
        '[:py:obj:`None`, :py:obj:`None`, '
        ':py:obj:`None`]'
    )
    assert restify(t.Iterator[None]) == (
        ':py:class:`~typing.Iterator`\\ [:py:obj:`None`]'
    )
    assert restify(abc.Iterator[None]) == (
        ':py:class:`collections.abc.Iterator`\\ [:py:obj:`None`]'
    )


def test_restify_Annotated() -> None:
    ann_rst = restify(Annotated[str, 'foo', 'bar'])
    assert ann_rst == (
        ":py:class:`~typing.Annotated`\\ [:py:class:`str`, 'foo', 'bar']"
    )
    ann_rst = restify(Annotated[str, 'foo', 'bar'], 'smart')
    assert ann_rst == (
        ":py:class:`~typing.Annotated`\\ [:py:class:`str`, 'foo', 'bar']"
    )
    ann_rst = restify(Annotated[float, Gt(-10.0)])
    assert ann_rst == (
        ':py:class:`~typing.Annotated`\\ [:py:class:`float`, :py:class:`tests.test_util.test_util_typing.Gt`\\ (gt=\\ -10.0)]'
    )
    ann_rst = restify(Annotated[float, Gt(-10.0)], 'smart')
    assert ann_rst == (
        ':py:class:`~typing.Annotated`\\ [:py:class:`float`, :py:class:`~tests.test_util.test_util_typing.Gt`\\ (gt=\\ -10.0)]'
    )


def test_restify_type_hints_Callable() -> None:
    assert restify(t.Callable) == ':py:class:`~typing.Callable`'
    assert restify(t.Callable[[str], int]) == (
        ':py:class:`~typing.Callable`\\ [[:py:class:`str`], :py:class:`int`]'
    )
    assert restify(t.Callable[..., int]) == (
        ':py:class:`~typing.Callable`\\ [[...], :py:class:`int`]'
    )
    assert restify(abc.Callable) == ':py:class:`collections.abc.Callable`'
    assert restify(abc.Callable[[str], int]) == (
        ':py:class:`collections.abc.Callable`\\ [[:py:class:`str`], :py:class:`int`]'
    )
    assert restify(abc.Callable[..., int]) == (
        ':py:class:`collections.abc.Callable`\\ [[...], :py:class:`int`]'
    )


def test_restify_type_hints_Union() -> None:
    assert restify(Union[int]) == ':py:class:`int`'
    assert restify(Union[int, str]) == ':py:class:`int` | :py:class:`str`'
    assert restify(Optional[int]) == ':py:class:`int` | :py:obj:`None`'

    assert restify(Union[str, None]) == ':py:class:`str` | :py:obj:`None`'
    assert restify(Union[None, str]) == ':py:obj:`None` | :py:class:`str`'
    assert restify(Optional[str]) == ':py:class:`str` | :py:obj:`None`'

    assert restify(Union[int, str, None]) == (
        ':py:class:`int` | :py:class:`str` | :py:obj:`None`'
    )
    assert restify(Optional[Union[int, str]]) in {
        ':py:class:`str` | :py:class:`int` | :py:obj:`None`',
        ':py:class:`int` | :py:class:`str` | :py:obj:`None`',
    }

    assert restify(Union[int, Integral]) == (
        ':py:class:`int` | :py:class:`numbers.Integral`'
    )
    assert restify(Union[int, Integral], 'smart') == (
        ':py:class:`int` | :py:class:`~numbers.Integral`'
    )

    assert restify(Union[MyClass1, MyClass2]) == (
        ':py:class:`tests.test_util.test_util_typing.MyClass1`'
        ' | :py:class:`tests.test_util.test_util_typing.<MyClass2>`'
    )
    assert restify(Union[MyClass1, MyClass2], 'smart') == (
        ':py:class:`~tests.test_util.test_util_typing.MyClass1`'
        ' | :py:class:`~tests.test_util.test_util_typing.<MyClass2>`'
    )

    assert restify(Optional[Union[MyClass1, MyClass2]]) == (
        ':py:class:`tests.test_util.test_util_typing.MyClass1`'
        ' | :py:class:`tests.test_util.test_util_typing.<MyClass2>`'
        ' | :py:obj:`None`'
    )
    assert restify(Optional[Union[MyClass1, MyClass2]], 'smart') == (
        ':py:class:`~tests.test_util.test_util_typing.MyClass1`'
        ' | :py:class:`~tests.test_util.test_util_typing.<MyClass2>`'
        ' | :py:obj:`None`'
    )


def test_restify_type_hints_typevars():
    T = TypeVar('T')
    T_co = TypeVar('T_co', covariant=True)
    T_contra = TypeVar('T_contra', contravariant=True)

    assert restify(T) == ':py:obj:`tests.test_util.test_util_typing.T`'
    assert restify(T, 'smart') == ':py:obj:`~tests.test_util.test_util_typing.T`'

    assert restify(T_co) == ':py:obj:`tests.test_util.test_util_typing.T_co`'
    assert restify(T_co, 'smart') == ':py:obj:`~tests.test_util.test_util_typing.T_co`'

    assert restify(T_contra) == ':py:obj:`tests.test_util.test_util_typing.T_contra`'
    ann_rst = restify(T_contra, 'smart')
    assert ann_rst == ':py:obj:`~tests.test_util.test_util_typing.T_contra`'

    ann_rst = restify(List[T])
    assert ann_rst == (
        ':py:class:`~typing.List`\\ [:py:obj:`tests.test_util.test_util_typing.T`]'
    )
    ann_rst = restify(List[T], 'smart')
    assert ann_rst == (
        ':py:class:`~typing.List`\\ [:py:obj:`~tests.test_util.test_util_typing.T`]'
    )

    ann_rst = restify(list[T])
    assert ann_rst == (
        ':py:class:`list`\\ [:py:obj:`tests.test_util.test_util_typing.T`]'
    )
    ann_rst = restify(list[T], 'smart')
    assert ann_rst == (
        ':py:class:`list`\\ [:py:obj:`~tests.test_util.test_util_typing.T`]'
    )

    assert restify(MyInt) == ':py:class:`tests.test_util.test_util_typing.MyInt`'
    ann_rst = restify(MyInt, 'smart')
    assert ann_rst == ':py:class:`~tests.test_util.test_util_typing.MyInt`'


def test_restify_type_hints_custom_class() -> None:
    assert restify(MyClass1) == ':py:class:`tests.test_util.test_util_typing.MyClass1`'
    ann_rst = restify(MyClass1, 'smart')
    assert ann_rst == ':py:class:`~tests.test_util.test_util_typing.MyClass1`'

    ann_rst = restify(MyClass2)
    assert ann_rst == ':py:class:`tests.test_util.test_util_typing.<MyClass2>`'
    ann_rst = restify(MyClass2, 'smart')
    assert ann_rst == ':py:class:`~tests.test_util.test_util_typing.<MyClass2>`'


def test_restify_type_hints_alias() -> None:
    MyStr = str
    MyTypingTuple = Tuple[str, str]
    MyTuple = tuple[str, str]
    assert restify(MyStr) == ':py:class:`str`'
    ann_rst = restify(MyTypingTuple)
    assert ann_rst == ':py:class:`~typing.Tuple`\\ [:py:class:`str`, :py:class:`str`]'
    assert restify(MyTuple) == ':py:class:`tuple`\\ [:py:class:`str`, :py:class:`str`]'


def test_restify_type_ForwardRef():
    assert restify(ForwardRef('MyInt')) == ':py:class:`MyInt`'

    assert (
        restify(list[ForwardRef('MyInt')]) == ':py:class:`list`\\ [:py:class:`MyInt`]'
    )

    ann_rst = restify(Tuple[dict[ForwardRef('MyInt'), str], list[List[int]]])
    assert ann_rst == (
        ':py:class:`~typing.Tuple`\\ [:py:class:`dict`\\ [:py:class:`MyInt`, :py:class:`str`], :py:class:`list`\\ [:py:class:`~typing.List`\\ [:py:class:`int`]]]'
    )


def test_restify_type_Literal() -> None:
    ann_rst = restify(Literal[1, '2', '\r'])
    assert ann_rst == ":py:obj:`~typing.Literal`\\ [1, '2', '\\r']"

    ann_rst = restify(Literal[MyEnum.a], 'fully-qualified-except-typing')
    assert ann_rst == (
        ':py:obj:`~typing.Literal`\\ [:py:attr:`tests.test_util.test_util_typing.MyEnum.a`]'
    )
    ann_rst = restify(Literal[MyEnum.a], 'smart')
    assert ann_rst == (
        ':py:obj:`~typing.Literal`\\ [:py:attr:`~tests.test_util.test_util_typing.MyEnum.a`]'
    )


def test_restify_pep_585() -> None:
    assert restify(list[str]) == ':py:class:`list`\\ [:py:class:`str`]'
    ann_rst = restify(dict[str, str])
    assert ann_rst == ':py:class:`dict`\\ [:py:class:`str`, :py:class:`str`]'
    assert restify(tuple[str, ...]) == ':py:class:`tuple`\\ [:py:class:`str`, ...]'
    assert restify(tuple[str, str, str]) == (
        ':py:class:`tuple`\\ [:py:class:`str`, :py:class:`str`, :py:class:`str`]'
    )
    ann_rst = restify(dict[str, tuple[int, ...]])
    assert ann_rst == (
        ':py:class:`dict`\\ '
        '[:py:class:`str`, :py:class:`tuple`\\ '
        '[:py:class:`int`, ...]]'
    )

    assert restify(tuple[()]) == ':py:class:`tuple`\\ [()]'

    # Mix old typing with PEP 585
    assert restify(List[dict[str, Tuple[str, ...]]]) == (
        ':py:class:`~typing.List`\\ '
        '[:py:class:`dict`\\ '
        '[:py:class:`str`, :py:class:`~typing.Tuple`\\ '
        '[:py:class:`str`, ...]]]'
    )
    assert restify(tuple[MyList[list[int]], int]) == (
        ':py:class:`tuple`\\ ['
        ':py:class:`tests.test_util.test_util_typing.MyList`\\ '
        '[:py:class:`list`\\ [:py:class:`int`]], '
        ':py:class:`int`]'
    )


def test_restify_Unpack() -> None:
    from typing_extensions import Unpack as UnpackCompat

    class X(t.TypedDict):
        x: int
        y: int
        label: str

    # Unpack is considered as typing special form so we always have '~'
    if sys.version_info[:2] >= (3, 12):
        expect = r':py:obj:`~typing.Unpack`\ [:py:class:`X`]'
        assert restify(UnpackCompat['X'], 'fully-qualified-except-typing') == expect
        assert restify(UnpackCompat['X'], 'smart') == expect
    else:
        expect = r':py:obj:`~typing_extensions.Unpack`\ [:py:class:`X`]'
        assert restify(UnpackCompat['X'], 'fully-qualified-except-typing') == expect
        assert restify(UnpackCompat['X'], 'smart') == expect

    expect = r':py:obj:`~typing.Unpack`\ [:py:class:`X`]'
    assert restify(t.Unpack['X'], 'fully-qualified-except-typing') == expect
    assert restify(t.Unpack['X'], 'smart') == expect


def test_restify_type_union_operator() -> None:
    assert restify(int | None) == ':py:class:`int` | :py:obj:`None`'
    assert restify(None | int) == ':py:obj:`None` | :py:class:`int`'
    assert restify(int | str) == ':py:class:`int` | :py:class:`str`'
    ann_rst = restify(int | str | None)
    assert ann_rst == ':py:class:`int` | :py:class:`str` | :py:obj:`None`'


def test_restify_broken_type_hints() -> None:
    ann_rst = restify(BrokenType)
    assert ann_rst == ':py:class:`tests.test_util.test_util_typing.BrokenType`'
    ann_rst = restify(BrokenType, 'smart')
    assert ann_rst == ':py:class:`~tests.test_util.test_util_typing.BrokenType`'


def test_restify_mock() -> None:
    with mock(['unknown']):
        import unknown  # type: ignore[import-not-found]

        assert restify(unknown) == ':py:class:`unknown`'
        assert restify(unknown.secret.Class) == ':py:class:`unknown.secret.Class`'
        ann_rst = restify(unknown.secret.Class, 'smart')
        assert ann_rst == ':py:class:`~unknown.secret.Class`'


def test_restify_type_hints_paramspec():
    P = ParamSpec('P')

    assert restify(P) == ':py:obj:`tests.test_util.test_util_typing.P`'
    assert restify(P, 'smart') == ':py:obj:`~tests.test_util.test_util_typing.P`'

    assert restify(P.args) == 'P.args'
    assert restify(P.args, 'smart') == 'P.args'

    assert restify(P.kwargs) == 'P.kwargs'
    assert restify(P.kwargs, 'smart') == 'P.kwargs'


def test_stringify_annotation() -> None:
    assert stringify_annotation(int, 'fully-qualified-except-typing') == 'int'
    assert stringify_annotation(int, 'smart') == 'int'

    assert stringify_annotation(str, 'fully-qualified-except-typing') == 'str'
    assert stringify_annotation(str, 'smart') == 'str'

    assert stringify_annotation(None, 'fully-qualified-except-typing') == 'None'
    assert stringify_annotation(None, 'smart') == 'None'

    ann_str = stringify_annotation(Integral, 'fully-qualified-except-typing')
    assert ann_str == 'numbers.Integral'
    assert stringify_annotation(Integral, 'smart') == '~numbers.Integral'

    ann_str = stringify_annotation(Struct, 'fully-qualified-except-typing')
    assert ann_str == 'struct.Struct'
    assert stringify_annotation(Struct, 'smart') == '~struct.Struct'

    ann_str = stringify_annotation(TracebackType, 'fully-qualified-except-typing')
    assert ann_str == 'types.TracebackType'
    assert stringify_annotation(TracebackType, 'smart') == '~types.TracebackType'

    ann_str = stringify_annotation(Path, 'fully-qualified-except-typing')
    assert ann_str == 'pathlib.Path'
    assert stringify_annotation(Path, 'smart') == '~pathlib.Path'

    assert stringify_annotation(Any, 'fully-qualified-except-typing') == 'Any'
    assert stringify_annotation(Any, 'fully-qualified') == 'typing.Any'
    assert stringify_annotation(Any, 'smart') == '~typing.Any'


def test_stringify_type_hints_containers():
    assert stringify_annotation(List, 'fully-qualified-except-typing') == 'List'
    assert stringify_annotation(List, 'fully-qualified') == 'typing.List'
    assert stringify_annotation(List, 'smart') == '~typing.List'

    assert stringify_annotation(Dict, 'fully-qualified-except-typing') == 'Dict'
    assert stringify_annotation(Dict, 'fully-qualified') == 'typing.Dict'
    assert stringify_annotation(Dict, 'smart') == '~typing.Dict'

    ann_str = stringify_annotation(List[int], 'fully-qualified-except-typing')
    assert ann_str == 'List[int]'
    assert stringify_annotation(List[int], 'fully-qualified') == 'typing.List[int]'
    assert stringify_annotation(List[int], 'smart') == '~typing.List[int]'

    ann_str = stringify_annotation(List[str], 'fully-qualified-except-typing')
    assert ann_str == 'List[str]'
    assert stringify_annotation(List[str], 'fully-qualified') == 'typing.List[str]'
    assert stringify_annotation(List[str], 'smart') == '~typing.List[str]'

    ann_str = stringify_annotation(Dict[str, float], 'fully-qualified-except-typing')
    assert ann_str == 'Dict[str, float]'
    ann_str = stringify_annotation(Dict[str, float], 'fully-qualified')
    assert ann_str == 'typing.Dict[str, float]'
    assert stringify_annotation(Dict[str, float], 'smart') == '~typing.Dict[str, float]'

    ann_str = stringify_annotation(
        Tuple[str, str, str], 'fully-qualified-except-typing'
    )
    assert ann_str == 'Tuple[str, str, str]'
    ann_str = stringify_annotation(Tuple[str, str, str], 'fully-qualified')
    assert ann_str == 'typing.Tuple[str, str, str]'
    ann_str = stringify_annotation(Tuple[str, str, str], 'smart')
    assert ann_str == '~typing.Tuple[str, str, str]'

    ann_str = stringify_annotation(Tuple[str, ...], 'fully-qualified-except-typing')
    assert ann_str == 'Tuple[str, ...]'
    ann_str = stringify_annotation(Tuple[str, ...], 'fully-qualified')
    assert ann_str == 'typing.Tuple[str, ...]'
    assert stringify_annotation(Tuple[str, ...], 'smart') == '~typing.Tuple[str, ...]'

    assert stringify_annotation(Tuple[()], 'fully-qualified-except-typing') == 'Tuple'
    assert stringify_annotation(Tuple[()], 'fully-qualified') == 'typing.Tuple'
    assert stringify_annotation(Tuple[()], 'smart') == '~typing.Tuple'

    ann_str = stringify_annotation(
        List[Dict[str, Tuple]], 'fully-qualified-except-typing'
    )
    assert ann_str == 'List[Dict[str, Tuple]]'
    ann_str = stringify_annotation(List[Dict[str, Tuple]], 'fully-qualified')
    assert ann_str == 'typing.List[typing.Dict[str, typing.Tuple]]'
    ann_str = stringify_annotation(List[Dict[str, Tuple]], 'smart')
    assert ann_str == '~typing.List[~typing.Dict[str, ~typing.Tuple]]'

    ann_str = stringify_annotation(
        MyList[Tuple[int, int]], 'fully-qualified-except-typing'
    )
    assert ann_str == 'tests.test_util.test_util_typing.MyList[Tuple[int, int]]'
    ann_str = stringify_annotation(MyList[Tuple[int, int]], 'fully-qualified')
    assert ann_str == (
        'tests.test_util.test_util_typing.MyList[typing.Tuple[int, int]]'
    )
    ann_str = stringify_annotation(MyList[Tuple[int, int]], 'smart')
    assert ann_str == (
        '~tests.test_util.test_util_typing.MyList[~typing.Tuple[int, int]]'
    )

    ann_str = stringify_annotation(
        t.Generator[None, None, None], 'fully-qualified-except-typing'
    )
    assert ann_str == 'Generator[None, None, None]'
    ann_str = stringify_annotation(t.Generator[None, None, None], 'fully-qualified')
    assert ann_str == 'typing.Generator[None, None, None]'
    ann_str = stringify_annotation(t.Generator[None, None, None], 'smart')
    assert ann_str == '~typing.Generator[None, None, None]'

    ann_str = stringify_annotation(
        abc.Generator[None, None, None], 'fully-qualified-except-typing'
    )
    assert ann_str == 'collections.abc.Generator[None, None, None]'
    ann_str = stringify_annotation(abc.Generator[None, None, None], 'fully-qualified')
    assert ann_str == 'collections.abc.Generator[None, None, None]'
    ann_str = stringify_annotation(abc.Generator[None, None, None], 'smart')
    assert ann_str == '~collections.abc.Generator[None, None, None]'

    ann_str = stringify_annotation(t.Iterator[None], 'fully-qualified-except-typing')
    assert ann_str == 'Iterator[None]'
    ann_str = stringify_annotation(t.Iterator[None], 'fully-qualified')
    assert ann_str == 'typing.Iterator[None]'
    assert stringify_annotation(t.Iterator[None], 'smart') == '~typing.Iterator[None]'

    ann_str = stringify_annotation(abc.Iterator[None], 'fully-qualified-except-typing')
    assert ann_str == 'collections.abc.Iterator[None]'
    ann_str = stringify_annotation(abc.Iterator[None], 'fully-qualified')
    assert ann_str == 'collections.abc.Iterator[None]'
    ann_str = stringify_annotation(abc.Iterator[None], 'smart')
    assert ann_str == '~collections.abc.Iterator[None]'


def test_stringify_type_hints_pep_585():
    ann_str = stringify_annotation(list[int], 'fully-qualified-except-typing')
    assert ann_str == 'list[int]'
    assert stringify_annotation(list[int], 'smart') == 'list[int]'

    ann_str = stringify_annotation(list[str], 'fully-qualified-except-typing')
    assert ann_str == 'list[str]'
    assert stringify_annotation(list[str], 'smart') == 'list[str]'

    ann_str = stringify_annotation(dict[str, float], 'fully-qualified-except-typing')
    assert ann_str == 'dict[str, float]'
    assert stringify_annotation(dict[str, float], 'smart') == 'dict[str, float]'

    ann_str = stringify_annotation(
        tuple[str, str, str], 'fully-qualified-except-typing'
    )
    assert ann_str == 'tuple[str, str, str]'
    assert stringify_annotation(tuple[str, str, str], 'smart') == 'tuple[str, str, str]'

    ann_str = stringify_annotation(tuple[str, ...], 'fully-qualified-except-typing')
    assert ann_str == 'tuple[str, ...]'
    assert stringify_annotation(tuple[str, ...], 'smart') == 'tuple[str, ...]'

    assert (
        stringify_annotation(tuple[()], 'fully-qualified-except-typing') == 'tuple[()]'
    )
    assert stringify_annotation(tuple[()], 'smart') == 'tuple[()]'

    ann_str = stringify_annotation(
        list[dict[str, tuple]], 'fully-qualified-except-typing'
    )
    assert ann_str == 'list[dict[str, tuple]]'
    ann_str = stringify_annotation(list[dict[str, tuple]], 'smart')
    assert ann_str == 'list[dict[str, tuple]]'

    ann_str = stringify_annotation(
        MyList[tuple[int, int]], 'fully-qualified-except-typing'
    )
    assert ann_str == 'tests.test_util.test_util_typing.MyList[tuple[int, int]]'
    ann_str = stringify_annotation(MyList[tuple[int, int]], 'fully-qualified')
    assert ann_str == 'tests.test_util.test_util_typing.MyList[tuple[int, int]]'
    ann_str = stringify_annotation(MyList[tuple[int, int]], 'smart')
    assert ann_str == '~tests.test_util.test_util_typing.MyList[tuple[int, int]]'

    ann_str = stringify_annotation(type[int], 'fully-qualified-except-typing')
    assert ann_str == 'type[int]'
    assert stringify_annotation(type[int], 'smart') == 'type[int]'

    # Mix typing and pep 585
    ann_str = stringify_annotation(
        tuple[List[dict[int, str]], str, ...], 'fully-qualified-except-typing'
    )
    assert ann_str == 'tuple[List[dict[int, str]], str, ...]'
    ann_str = stringify_annotation(tuple[List[dict[int, str]], str, ...], 'smart')
    assert ann_str == 'tuple[~typing.List[dict[int, str]], str, ...]'


def test_stringify_Annotated() -> None:
    ann_str = stringify_annotation(
        Annotated[str, 'foo', 'bar'], 'fully-qualified-except-typing'
    )
    assert ann_str == "Annotated[str, 'foo', 'bar']"
    ann_str = stringify_annotation(Annotated[str, 'foo', 'bar'], 'smart')
    assert ann_str == "~typing.Annotated[str, 'foo', 'bar']"
    ann_str = stringify_annotation(
        Annotated[float, Gt(-10.0)], 'fully-qualified-except-typing'
    )
    assert ann_str == (
        'Annotated[float, tests.test_util.test_util_typing.Gt(gt=-10.0)]'
    )
    ann_str = stringify_annotation(Annotated[float, Gt(-10.0)], 'smart')
    assert ann_str == (
        '~typing.Annotated[float, ~tests.test_util.test_util_typing.Gt(gt=-10.0)]'
    )


def test_stringify_Unpack() -> None:
    class X(t.TypedDict):
        x: int
        y: int
        label: str

    assert stringify_annotation(t.Unpack['X']) == 'Unpack[X]'
    assert stringify_annotation(t.Unpack['X'], 'smart') == '~typing.Unpack[X]'


def test_stringify_type_hints_string() -> None:
    assert stringify_annotation('int', 'fully-qualified-except-typing') == 'int'
    assert stringify_annotation('int', 'fully-qualified') == 'int'
    assert stringify_annotation('int', 'smart') == 'int'

    assert stringify_annotation('str', 'fully-qualified-except-typing') == 'str'
    assert stringify_annotation('str', 'fully-qualified') == 'str'
    assert stringify_annotation('str', 'smart') == 'str'

    ann_str = stringify_annotation(List['int'], 'fully-qualified-except-typing')
    assert ann_str == 'List[int]'
    assert stringify_annotation(List['int'], 'fully-qualified') == 'typing.List[int]'
    assert stringify_annotation(List['int'], 'smart') == '~typing.List[int]'

    ann_str = stringify_annotation(list['int'], 'fully-qualified-except-typing')
    assert ann_str == 'list[int]'
    assert stringify_annotation(list['int'], 'fully-qualified') == 'list[int]'
    assert stringify_annotation(list['int'], 'smart') == 'list[int]'

    ann_str = stringify_annotation('Tuple[str]', 'fully-qualified-except-typing')
    assert ann_str == 'Tuple[str]'
    assert stringify_annotation('Tuple[str]', 'fully-qualified') == 'Tuple[str]'
    assert stringify_annotation('Tuple[str]', 'smart') == 'Tuple[str]'

    ann_str = stringify_annotation('tuple[str]', 'fully-qualified-except-typing')
    assert ann_str == 'tuple[str]'
    assert stringify_annotation('tuple[str]', 'fully-qualified') == 'tuple[str]'
    assert stringify_annotation('tuple[str]', 'smart') == 'tuple[str]'

    assert stringify_annotation('unknown', 'fully-qualified-except-typing') == 'unknown'
    assert stringify_annotation('unknown', 'fully-qualified') == 'unknown'
    assert stringify_annotation('unknown', 'smart') == 'unknown'


def test_stringify_type_hints_Callable() -> None:
    ann_str = stringify_annotation(t.Callable, 'fully-qualified-except-typing')
    assert ann_str == 'Callable'
    assert stringify_annotation(t.Callable, 'fully-qualified') == 'typing.Callable'
    assert stringify_annotation(t.Callable, 'smart') == '~typing.Callable'

    ann_str = stringify_annotation(
        t.Callable[[str], int], 'fully-qualified-except-typing'
    )
    assert ann_str == 'Callable[[str], int]'
    ann_str = stringify_annotation(t.Callable[[str], int], 'fully-qualified')
    assert ann_str == 'typing.Callable[[str], int]'
    ann_str = stringify_annotation(t.Callable[[str], int], 'smart')
    assert ann_str == '~typing.Callable[[str], int]'

    ann_str = stringify_annotation(
        t.Callable[..., int], 'fully-qualified-except-typing'
    )
    assert ann_str == 'Callable[[...], int]'
    ann_str = stringify_annotation(t.Callable[..., int], 'fully-qualified')
    assert ann_str == 'typing.Callable[[...], int]'
    ann_str = stringify_annotation(t.Callable[..., int], 'smart')
    assert ann_str == '~typing.Callable[[...], int]'

    ann_str = stringify_annotation(abc.Callable, 'fully-qualified-except-typing')
    assert ann_str == 'collections.abc.Callable'
    ann_str = stringify_annotation(abc.Callable, 'fully-qualified')
    assert ann_str == 'collections.abc.Callable'
    assert stringify_annotation(abc.Callable, 'smart') == '~collections.abc.Callable'

    ann_str = stringify_annotation(
        abc.Callable[[str], int], 'fully-qualified-except-typing'
    )
    assert ann_str == 'collections.abc.Callable[[str], int]'
    ann_str = stringify_annotation(abc.Callable[[str], int], 'fully-qualified')
    assert ann_str == 'collections.abc.Callable[[str], int]'
    ann_str = stringify_annotation(abc.Callable[[str], int], 'smart')
    assert ann_str == '~collections.abc.Callable[[str], int]'

    ann_str = stringify_annotation(
        abc.Callable[..., int], 'fully-qualified-except-typing'
    )
    assert ann_str == 'collections.abc.Callable[[...], int]'
    ann_str = stringify_annotation(abc.Callable[..., int], 'fully-qualified')
    assert ann_str == 'collections.abc.Callable[[...], int]'
    ann_str = stringify_annotation(abc.Callable[..., int], 'smart')
    assert ann_str == '~collections.abc.Callable[[...], int]'


def test_stringify_type_hints_Union() -> None:
    ann_str = stringify_annotation(Optional[int], 'fully-qualified-except-typing')
    assert ann_str == 'int | None'
    assert stringify_annotation(Optional[int], 'fully-qualified') == 'int | None'
    assert stringify_annotation(Optional[int], 'smart') == 'int | None'

    ann_str = stringify_annotation(Union[int, None], 'fully-qualified-except-typing')
    assert ann_str == 'int | None'
    ann_str = stringify_annotation(Union[None, int], 'fully-qualified-except-typing')
    assert ann_str == 'None | int'
    assert stringify_annotation(Union[int, None], 'fully-qualified') == 'int | None'
    assert stringify_annotation(Union[None, int], 'fully-qualified') == 'None | int'
    assert stringify_annotation(Union[int, None], 'smart') == 'int | None'
    assert stringify_annotation(Union[None, int], 'smart') == 'None | int'

    ann_str = stringify_annotation(Union[int, str], 'fully-qualified-except-typing')
    assert ann_str == 'int | str'
    assert stringify_annotation(Union[int, str], 'fully-qualified') == 'int | str'
    assert stringify_annotation(Union[int, str], 'smart') == 'int | str'

    ann_str = stringify_annotation(
        Union[int, Integral], 'fully-qualified-except-typing'
    )
    assert ann_str == 'int | numbers.Integral'
    ann_str = stringify_annotation(Union[int, Integral], 'fully-qualified')
    assert ann_str == 'int | numbers.Integral'
    ann_str = stringify_annotation(Union[int, Integral], 'smart')
    assert ann_str == 'int | ~numbers.Integral'

    ann_str = stringify_annotation(
        Union[MyClass1, MyClass2], 'fully-qualified-except-typing'
    )
    assert ann_str == (
        'tests.test_util.test_util_typing.MyClass1 | tests.test_util.test_util_typing.<MyClass2>'
    )
    ann_str = stringify_annotation(Union[MyClass1, MyClass2], 'fully-qualified')
    assert ann_str == (
        'tests.test_util.test_util_typing.MyClass1 | tests.test_util.test_util_typing.<MyClass2>'
    )
    ann_str = stringify_annotation(Union[MyClass1, MyClass2], 'smart')
    assert ann_str == (
        '~tests.test_util.test_util_typing.MyClass1 | ~tests.test_util.test_util_typing.<MyClass2>'
    )


def test_stringify_type_hints_typevars():
    T = TypeVar('T')
    T_co = TypeVar('T_co', covariant=True)
    T_contra = TypeVar('T_contra', contravariant=True)

    ann_str = stringify_annotation(T, 'fully-qualified-except-typing')
    assert ann_str == 'tests.test_util.test_util_typing.T'
    assert stringify_annotation(T, 'smart') == '~tests.test_util.test_util_typing.T'

    ann_str = stringify_annotation(T_co, 'fully-qualified-except-typing')
    assert ann_str == 'tests.test_util.test_util_typing.T_co'
    ann_str = stringify_annotation(T_co, 'smart')
    assert ann_str == '~tests.test_util.test_util_typing.T_co'

    ann_str = stringify_annotation(T_contra, 'fully-qualified-except-typing')
    assert ann_str == 'tests.test_util.test_util_typing.T_contra'
    ann_str = stringify_annotation(T_contra, 'smart')
    assert ann_str == '~tests.test_util.test_util_typing.T_contra'

    ann_str = stringify_annotation(List[T], 'fully-qualified-except-typing')
    assert ann_str == 'List[tests.test_util.test_util_typing.T]'
    ann_str = stringify_annotation(List[T], 'smart')
    assert ann_str == '~typing.List[~tests.test_util.test_util_typing.T]'

    ann_str = stringify_annotation(list[T], 'fully-qualified-except-typing')
    assert ann_str == 'list[tests.test_util.test_util_typing.T]'
    ann_str = stringify_annotation(list[T], 'smart')
    assert ann_str == 'list[~tests.test_util.test_util_typing.T]'

    ann_str = stringify_annotation(MyInt, 'fully-qualified-except-typing')
    assert ann_str == 'tests.test_util.test_util_typing.MyInt'
    ann_str = stringify_annotation(MyInt, 'smart')
    assert ann_str == '~tests.test_util.test_util_typing.MyInt'


def test_stringify_type_hints_custom_class() -> None:
    ann_str = stringify_annotation(MyClass1, 'fully-qualified-except-typing')
    assert ann_str == 'tests.test_util.test_util_typing.MyClass1'
    ann_str = stringify_annotation(MyClass1, 'smart')
    assert ann_str == '~tests.test_util.test_util_typing.MyClass1'

    ann_str = stringify_annotation(MyClass2, 'fully-qualified-except-typing')
    assert ann_str == 'tests.test_util.test_util_typing.<MyClass2>'
    ann_str = stringify_annotation(MyClass2, 'smart')
    assert ann_str == '~tests.test_util.test_util_typing.<MyClass2>'


def test_stringify_type_hints_alias() -> None:
    MyStr = str
    MyTuple = Tuple[str, str]

    assert stringify_annotation(MyStr, 'fully-qualified-except-typing') == 'str'
    assert stringify_annotation(MyStr, 'smart') == 'str'

    assert stringify_annotation(MyTuple) == 'Tuple[str, str]'
    assert stringify_annotation(MyTuple, 'smart') == '~typing.Tuple[str, str]'


def test_stringify_type_Literal() -> None:
    ann_str = stringify_annotation(
        Literal[1, '2', '\r'], 'fully-qualified-except-typing'
    )
    assert ann_str == "Literal[1, '2', '\\r']"
    ann_str = stringify_annotation(Literal[1, '2', '\r'], 'fully-qualified')
    assert ann_str == "typing.Literal[1, '2', '\\r']"
    ann_str = stringify_annotation(Literal[1, '2', '\r'], 'smart')
    assert ann_str == "~typing.Literal[1, '2', '\\r']"

    ann_str = stringify_annotation(Literal[MyEnum.a], 'fully-qualified-except-typing')
    assert ann_str == 'Literal[tests.test_util.test_util_typing.MyEnum.a]'
    ann_str = stringify_annotation(Literal[MyEnum.a], 'fully-qualified')
    assert ann_str == 'typing.Literal[tests.test_util.test_util_typing.MyEnum.a]'
    ann_str = stringify_annotation(Literal[MyEnum.a], 'smart')
    assert ann_str == '~typing.Literal[MyEnum.a]'


def test_stringify_type_union_operator() -> None:
    assert stringify_annotation(int | None) == 'int | None'
    assert stringify_annotation(int | None, 'smart') == 'int | None'

    assert stringify_annotation(int | str) == 'int | str'
    assert stringify_annotation(int | str, 'smart') == 'int | str'

    assert stringify_annotation(int | str | None) == 'int | str | None'
    assert stringify_annotation(int | str | None, 'smart') == 'int | str | None'

    ann_str = stringify_annotation(
        int | tuple[dict[str, int | None], list[int | str]] | None
    )
    assert ann_str == 'int | tuple[dict[str, int | None], list[int | str]] | None'
    ann_str = stringify_annotation(
        int | tuple[dict[str, int | None], list[int | str]] | None, 'smart'
    )
    assert ann_str == 'int | tuple[dict[str, int | None], list[int | str]] | None'

    assert stringify_annotation(int | Struct) == 'int | struct.Struct'
    assert stringify_annotation(int | Struct, 'smart') == 'int | ~struct.Struct'


def test_stringify_broken_type_hints() -> None:
    ann_str = stringify_annotation(BrokenType, 'fully-qualified-except-typing')
    assert ann_str == 'tests.test_util.test_util_typing.BrokenType'
    ann_str = stringify_annotation(BrokenType, 'smart')
    assert ann_str == '~tests.test_util.test_util_typing.BrokenType'


def test_stringify_mock() -> None:
    with mock(['unknown']):
        import unknown

        ann_str = stringify_annotation(unknown, 'fully-qualified-except-typing')
        assert ann_str == 'unknown'
        ann_str = stringify_annotation(
            unknown.secret.Class, 'fully-qualified-except-typing'
        )
        assert ann_str == 'unknown.secret.Class'
        ann_str = stringify_annotation(unknown.secret.Class, 'smart')
        assert ann_str == 'unknown.secret.Class'


def test_stringify_type_ForwardRef():
    assert stringify_annotation(ForwardRef('MyInt')) == 'MyInt'
    assert stringify_annotation(ForwardRef('MyInt'), 'smart') == 'MyInt'

    assert stringify_annotation(list[ForwardRef('MyInt')]) == 'list[MyInt]'
    assert stringify_annotation(list[ForwardRef('MyInt')], 'smart') == 'list[MyInt]'

    ann_str = stringify_annotation(
        Tuple[dict[ForwardRef('MyInt'), str], list[List[int]]]
    )
    assert ann_str == 'Tuple[dict[MyInt, str], list[List[int]]]'
    ann_str = stringify_annotation(
        Tuple[dict[ForwardRef('MyInt'), str], list[List[int]]],
        'fully-qualified-except-typing',
    )
    assert ann_str == 'Tuple[dict[MyInt, str], list[List[int]]]'
    ann_str = stringify_annotation(
        Tuple[dict[ForwardRef('MyInt'), str], list[List[int]]], 'smart'
    )
    assert ann_str == '~typing.Tuple[dict[MyInt, str], list[~typing.List[int]]]'


def test_stringify_type_hints_paramspec():
    P = ParamSpec('P')

    assert stringify_annotation(P, 'fully-qualified') == '~P'
    assert stringify_annotation(P, 'fully-qualified-except-typing') == '~P'
    assert stringify_annotation(P, 'smart') == '~P'

    assert stringify_annotation(P.args, 'fully-qualified') == 'typing.~P'
    assert stringify_annotation(P.args, 'fully-qualified-except-typing') == '~P'
    assert stringify_annotation(P.args, 'smart') == '~typing.~P'

    assert stringify_annotation(P.kwargs, 'fully-qualified') == 'typing.~P'
    assert stringify_annotation(P.kwargs, 'fully-qualified-except-typing') == '~P'
    assert stringify_annotation(P.kwargs, 'smart') == '~typing.~P'



"""Test sphinx.ext.githubpages extension."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sphinx.testing.util import SphinxTestApp

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='ext-githubpages')
def test_githubpages(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    assert (app.outdir / '.nojekyll').exists()
    assert not (app.outdir / 'CNAME').exists()


@pytest.mark.sphinx(
    'html',
    testroot='ext-githubpages',
    confoverrides={'html_baseurl': 'https://sphinx-doc.github.io'},
)
def test_no_cname_for_github_io_domain(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    assert (app.outdir / '.nojekyll').exists()
    assert not (app.outdir / 'CNAME').exists()


@pytest.mark.sphinx(
    'html',
    testroot='ext-githubpages',
    confoverrides={'html_baseurl': 'https://sphinx-doc.org'},
)
def test_cname_for_custom_domain(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    assert (app.outdir / '.nojekyll').exists()
    assert (app.outdir / 'CNAME').read_text(encoding='utf8') == 'sphinx-doc.org'



from .autosummary_dummy_module import Bar, PublicBar, foo, public_foo


def baz():
    """Baz function"""
    pass


def public_baz():
    """Public Baz function"""


__all__ = ['PublicBar', 'public_foo', 'public_baz', 'extra_dummy_module']  # NoQA: F822



"""Test the Theme class."""

from __future__ import annotations

import shutil
from pathlib import Path
from typing import TYPE_CHECKING
from xml.etree.ElementTree import ParseError

import pytest
from defusedxml.ElementTree import parse as xml_parse

import sphinx.builders.html
from sphinx.builders.html import StandaloneHTMLBuilder
from sphinx.errors import ThemeError
from sphinx.theming import (
    _ConfigFile,
    _convert_theme_conf,
    _convert_theme_toml,
    _load_theme,
    _load_theme_conf,
    _load_theme_toml,
)

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp

HERE = Path(__file__).resolve().parent


@pytest.mark.sphinx(
    'html',
    testroot='theming',
    confoverrides={'html_theme': 'ziptheme', 'html_theme_options.testopt': 'foo'},
)
def test_theme_api(app: SphinxTestApp) -> None:
    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    themes = [
        'basic',
        'default',
        'scrolls',
        'agogo',
        'sphinxdoc',
        'haiku',
        'traditional',
        'epub',
        'nature',
        'pyramid',
        'bizstyle',
        'classic',
        'nonav',
        'test-theme',
        'ziptheme',
        'staticfiles',
        'parent',
        'child',
        'alabaster',
    ]

    # test Theme class API
    assert set(app.registry.html_themes.keys()) == set(themes)
    assert app.registry.html_themes['test-theme'] == (
        app.srcdir / 'test_theme' / 'test-theme'
    )
    assert app.registry.html_themes['ziptheme'] == (app.srcdir / 'ziptheme.zip')
    assert app.registry.html_themes['staticfiles'] == (
        app.srcdir / 'test_theme' / 'staticfiles'
    )

    # test Theme instance API
    theme = app.builder.theme
    assert theme.name == 'ziptheme'
    assert len(theme.get_theme_dirs()) == 2

    # direct setting
    assert theme.get_config('theme', 'stylesheet') == 'custom.css'
    # inherited setting
    assert theme.get_config('options', 'nosidebar') == 'false'
    # nonexisting setting
    assert theme.get_config('theme', 'foobar', 'def') == 'def'
    with pytest.raises(ThemeError):
        theme.get_config('theme', 'foobar')
    # nonexisting section
    with pytest.raises(ThemeError):
        theme.get_config('foobar', 'foobar')

    # options API

    options = theme.get_options({'nonexisting': 'foo'})
    assert 'nonexisting' not in options

    options = theme.get_options(app.config.html_theme_options)
    assert options['testopt'] == 'foo'
    assert options['nosidebar'] == 'false'

    # cleanup temp directories
    theme._cleanup()
    assert not any(p.exists() for p in theme._tmp_dirs)


def test_nonexistent_theme_settings(tmp_path):
    # Check that error occurs with a non-existent theme.toml or theme.conf
    # https://github.com/sphinx-doc/sphinx/issues/11668
    with pytest.raises(ThemeError):
        _load_theme('', tmp_path)


@pytest.mark.sphinx('html', testroot='double-inheriting-theme')
def test_double_inheriting_theme(app: SphinxTestApp) -> None:
    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    assert app.builder.theme.name == 'base_theme2'
    app.build()  # => not raises TemplateNotFound


@pytest.mark.sphinx(
    'html',
    testroot='theming',
    confoverrides={'html_theme': 'child'},
)
def test_nested_zipped_theme(app: SphinxTestApp) -> None:
    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    assert app.builder.theme.name == 'child'
    app.build()  # => not raises TemplateNotFound


@pytest.mark.sphinx(
    'html',
    testroot='theming',
    confoverrides={'html_theme': 'staticfiles'},
)
def test_staticfiles(app: SphinxTestApp) -> None:
    app.build()
    assert (app.outdir / '_static' / 'legacytmpl.html').exists()
    assert (app.outdir / '_static' / 'legacytmpl.html').read_text(encoding='utf8') == (
        '<!-- testing legacy _t static templates -->\n'
        '<html><project>project name not set</project></html>'
    )
    assert (app.outdir / '_static' / 'staticimg.png').exists()
    assert (app.outdir / '_static' / 'statictmpl.html').exists()
    assert (app.outdir / '_static' / 'statictmpl.html').read_text(encoding='utf8') == (
        '<!-- testing static templates -->\n<html><project>Project name not set</project></html>'
    )

    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<meta name="testopt" content="optdefault" />' in result


@pytest.mark.sphinx(
    'html',
    testroot='theming',
    confoverrides={'html_theme': 'test-theme'},
)
def test_dark_style(app, monkeypatch):
    monkeypatch.setattr(sphinx.builders.html, '_file_checksum', lambda o, f: '')

    style = app.builder.dark_highlighter.formatter_args.get('style')
    assert style.__name__ == 'MonokaiStyle'

    app.build()
    assert (app.outdir / '_static' / 'pygments_dark.css').exists()

    css_file, properties = app.registry.css_files[0]
    assert css_file == 'pygments_dark.css'
    assert 'media' in properties
    assert properties['media'] == '(prefers-color-scheme: dark)'

    assert sorted(f.filename for f in app.builder._css_files) == [
        '_static/classic.css',
        '_static/pygments.css',
        '_static/pygments_dark.css',
    ]

    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert (
        '<link rel="stylesheet" type="text/css" href="_static/pygments.css" />'
    ) in result
    assert (
        '<link id="pygments_dark_css" media="(prefers-color-scheme: dark)" '
        'rel="stylesheet" type="text/css" '
        'href="_static/pygments_dark.css" />'
    ) in result


@pytest.mark.sphinx('html', testroot='theming')
def test_theme_sidebars(app: SphinxTestApp) -> None:
    app.build()

    # test-theme specifies globaltoc and searchbox as default sidebars
    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<h3><a href="#">Table of Contents</a></h3>' in result
    assert '<h3>Related Topics</h3>' not in result
    assert '<h3>This Page</h3>' not in result
    assert '<h3 id="searchlabel">Quick search</h3>' in result


@pytest.mark.parametrize(
    'theme_name',
    [
        'alabaster',
        'agogo',
        'basic',
        'bizstyle',
        'classic',
        'default',
        'epub',
        'haiku',
        'nature',
        'nonav',
        'pyramid',
        'scrolls',
        'sphinxdoc',
        'traditional',
    ],
)
def test_theme_builds(make_app, rootdir, sphinx_test_tempdir, theme_name):
    """Test all the themes included with Sphinx build a simple project and produce valid XML."""
    testroot_path = rootdir / 'test-basic'
    srcdir = sphinx_test_tempdir / f'test-theme-{theme_name}'
    shutil.copytree(testroot_path, srcdir)

    app = make_app(srcdir=srcdir, confoverrides={'html_theme': theme_name})
    app.build()
    assert not app.warning.getvalue().strip()
    assert app.outdir.joinpath('index.html').exists()

    # check that the generated HTML files are well-formed (as strict XML)
    for html_file in app.outdir.rglob('*.html'):
        try:
            xml_parse(html_file)
        except ParseError as exc:
            pytest.fail(f'Failed to parse {html_file.relative_to(app.outdir)}: {exc}')


def test_config_file_toml() -> None:
    config_path = HERE / 'theme.toml'
    cfg = _load_theme_toml(config_path)
    config = _convert_theme_toml(cfg)

    assert config == _ConfigFile(
        stylesheets=('spam.css', 'ham.css'),
        sidebar_templates=None,
        pygments_style_default='spam',
        pygments_style_dark=None,
        options={'lobster': 'thermidor'},
    )


def test_config_file_conf() -> None:
    config_path = HERE / 'theme.conf'
    cfg = _load_theme_conf(config_path)
    config = _convert_theme_conf(cfg)

    assert config == _ConfigFile(
        stylesheets=('spam.css', 'ham.css'),
        sidebar_templates=None,
        pygments_style_default='spam',
        pygments_style_dark=None,
        options={'lobster': 'thermidor'},
    )
